import subprocess
import sys
from typing import Any


def config_xmake(args):
    subprocess.run(["xmake", "f",
                    "-a", 'x86' if sys.maxsize < 2 ** 32 else 'x64']
                   + args)


if __name__ == "__main__":
    config_xmake(sys.argv[1:])

try:
    from hatchling.builders.hooks.plugin.interface import BuildHookInterface  # noqa
except ImportError:
    sys.exit()


class CustomBuildHook(BuildHookInterface):
    def clean(self, versions: list[str]):  # noqa
        subprocess.run("xmake c")

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:  # noqa
        platform = "win32" if sys.maxsize < 2 ** 32 else "win_amd64"
        pyver = "cp" + str(sys.version_info.major) + str(sys.version_info.minor)
        build_data["tag"] = f"{pyver}-{pyver}-{platform}"
        build_data["pure_python"] = False
        config_xmake(["-m", "release", "-c", "-y"])
        subprocess.run("xmake -r")

    def dependencies(self):  # noqa
        try:
            subprocess.run("xmake --version", stdout=subprocess.DEVNULL, check=True)
        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            raise RuntimeError("xmake is not installed") from e
        return []



# -*- coding: utf_8 -*-
"""
注入, 打开游戏相关的函数和类.
"""
import os
import signal
import subprocess
import time
from collections.abc import Iterable
from contextlib import ContextDecorator, AbstractContextManager
from typing import Self, overload

from . import asm
from .exception import PvzStatusError
from ..rp_extend import Controller, ControllerError


def open_game(game_path: str, num: int = 1) -> list[int]:
    """
    通过路径, 将 pvz 作为子进程打开游戏
    
    Args:
        game_path: 游戏路径, 绝对相对路径均可
        num: 打开的游戏数量
    Returns:
        打开的所有游戏进程 process id 组成的列表
    """
    abs_path = os.path.abspath(game_path)
    route, exe_name = os.path.split(abs_path)
    current_directory = os.getcwd()
    ret = [0] * num
    try:
        os.chdir(route)
        for i in range(num):
            process = subprocess.Popen(f"\"{exe_name}\"")
            ret[i] = process.pid
    finally:
        os.chdir(current_directory)
    return ret


def inject(pids: Iterable[int],
           stdout=subprocess.DEVNULL,
           set_console: bool = True) -> list[Controller]:
    """
    对 pids 中的每一个进程注入 .dll
    
    Args:
        pids: 所有 process id
        stdout: inject程序标准输出流, 默认丢弃
        set_console: 是否打开游戏控制台
    Returns:
        所有进程的 Controller 对象组成的列表
    """
    current_dir = os.getcwd()
    try:
        os.chdir(os.path.dirname(__file__))
        dll_path = os.path.abspath("..\\bin\\rp_dll.dll")
        s = f'..\\bin\\rp_injector.exe {1 if set_console else 0} \"{dll_path}\" '
        s += ' '.join(str(i) for i in pids)
        subprocess.run(s, stdout=stdout)
    finally:
        os.chdir(current_dir)
    return [Controller(pid) for pid in pids]


def close_by_pids(pids: Iterable[int]) -> None:
    """
    通过 process id 关闭进程

    Args:
        pids: 需要关闭的 process id
    """
    for pid in pids:
        os.kill(pid, signal.SIGTERM)


class InjectedGame(AbstractContextManager):
    """
    描述被注入游戏的类

    Attributes:
        controller: 被注入游戏的控制器
    """
    @overload
    def __init__(self, process_id: int, /, close_when_exit: bool = True):
        """
        通过已经注入的 process id 构造 InjectedGame 对象

        Args:
            process_id: pvz 进程的 process id
            close_when_exit: 是否在退出时关闭 pvz 进程
        """

    @overload
    def __init__(self, game_path: str, /, close_when_exit: bool = True):
        """
        通过游戏路径构造 InjectedGame 对象

        Args:
            game_path: pvz 主程序路径
            close_when_exit: 是否在退出时关闭 pvz 进程
        """

    @overload
    def __init__(self, controller: Controller, /, close_when_exit: bool = True):
        """
        通过 Controller 对象构造 InjectedGame 对象

        Args:
            controller: 注入目标游戏的 Controller 对象
            close_when_exit: 是否在退出时关闭 pvz 进程
        """

    def __init__(self, arg, /, close_when_exit: bool = True):
        self._close_when_exit = close_when_exit
        if isinstance(arg, int):
            self.controller: Controller = Controller(arg)
        elif isinstance(arg, str):
            self.controller: Controller = inject(open_game(arg))[0]
        elif isinstance(arg, Controller):
            self.controller: Controller = arg
        else:
            raise TypeError("the parameter should be int, str or Controller instance")
        
    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ControllerError:
            return
        self.controller.end()
        if (self._close_when_exit and
                exc_type is not KeyboardInterrupt):
            close_by_pids((self.controller.pid,))


class ConnectedContext(ContextDecorator):
    """
    创造已连接游戏的上下文

     Attributes:
         controller: 被注入游戏的控制器
         ensure_jump_frame: 是否保证跳帧, True 则保证跳帧, False 则保证不跳帧. 默认 None 不处理.
    """
    def __init__(self, controller: Controller, ensure_jump_frame: bool | None = None):
        self.controller: Controller = controller
        self.ensure_jump_frame = ensure_jump_frame
        self._is_connected: bool = False
        self._is_jumping: bool = False

    def __enter__(self) -> Controller:
        self._is_connected = self.controller.hook_connected()
        ctler = self.controller
        if not self._is_connected:
            ctler.start()
        if self.ensure_jump_frame is not None:
            self._is_jumping = ctler.is_jumping_frame()
            if self.ensure_jump_frame:
                ctler.start_jump_frame()
            else:
                ctler.end_jump_frame()
        return ctler

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ControllerError:
            return
        ctler = self.controller
        if self.ensure_jump_frame is not None:
            if self._is_jumping:
                ctler.start_jump_frame()
            else:
                ctler.end_jump_frame()
        if not self._is_connected:
            ctler.end()
        else:
            ctler.start()


def enter_level(controller: Controller, game_mode: int, look_for_saved_game: bool = False) -> None:
    """
    进入游戏关卡

    Args:
        controller: 目标游戏的 Controller
        game_mode: 关卡对应数字
        look_for_saved_game: 是否尝试读档, **请切记默认情况会毁坏你原有的关卡存档!**
    Raises:
        PvzStatusError: 若不在载入界面, 主界面, 游戏中或小游戏选项卡界面使用此函数则抛出
        ControllerError: 若 Controller 对象未连接游戏则抛出
    """
    code = f"""
        push esi
        mov esi, [{0x6a9ec0}]
        mov eax, [esi + {0x7fc}]
        test eax, eax
        jz LCompleteLoading
        cmp eax, 1  // main screen
        je LDeleteGameSelector
        cmp eax, 7  // challenge selector screen
        je LDeleteChallengeScreen
        mov edx, [esi + {0x768}]
        test edx, edx  // have Board
        jnz LNewBoard
        LError:
        mov [{controller.result_address}], eax
        pop esi
        ret
        
        LDeleteChallengeScreen:
        call {0x44fd00}  // LawnApp::KillChallengeScreen(esi = LawnApp* this)
        jmp LPreNewGame
        
        LNewBoard:
        mov cl, [edx + {0x5760}]
        mov [esi + {0x88c}], cl  // deal with yeti
        jmp LPreNewGame
        
        LCompleteLoading:
        mov ecx, esi
        call {0x452cb0}  // LawnApp::LoadingCompleted(ecx = LawnApp* this)
        
        LDeleteGameSelector:
        call {0x44f9e0}  // LawnApp::KillGameSelector(esi = LawnApp* this)
        
        LPreNewGame:
        push {int(look_for_saved_game)}
        push {game_mode}
        call {0x44f560}  // LawnApp::PreNewGame
        xor eax, eax;
        mov [{controller.result_address}], eax;
        pop esi;
        ret;"""
    with ConnectedContext(controller, False) as ctler:
        if ctler.read_u32(0x6a9ec0, 0x76c) is not None:
            ctler.end()
            while not ctler.read_bool(0x6a9ec0, 0x76c, 0xa1):  # 是否加载成功bool, thanks for ghast
                if not ctler.global_connected():
                    raise ControllerError("global hook not connected")
                time.sleep(0.1)
            ctler.start()
        asm.run(code, ctler)
        ctler.skip_frames()
    if controller.result_i32:
        raise PvzStatusError("this function should be used at loading screen, "
                             "main selector screen, challenge selector screen or in the game"
                             f"while the current screen num is {controller.result_i32}")



# -*- coding: utf_8 -*-
"""
ize 常数, 常见表
"""
from enum import Enum, auto

from ..structs.plant import PlantType
from ..structs.zombie import ZombieType

ize_plant_types: set[PlantType] = {
    PlantType.pea_shooter,
    PlantType.sunflower,
    PlantType.wallnut,
    PlantType.potato_mine,
    PlantType.snow_pea,
    PlantType.chomper,
    PlantType.repeater,
    PlantType.puffshroom,
    PlantType.fumeshroom,
    PlantType.scaredyshroom,
    PlantType.squash,
    PlantType.threepeater,
    PlantType.spikeweed,
    PlantType.torchwood,
    PlantType.split_pea,
    PlantType.starfruit,
    PlantType.magnetshroom,
    PlantType.kernelpult,
    PlantType.umbrella_leaf
}
"""所有 ize 中出现的植物"""

ize_zombie_types: set[ZombieType] = {
    ZombieType.imp,
    ZombieType.conehead,
    ZombieType.pole_vaulting,
    ZombieType.buckethead,
    ZombieType.bungee,
    ZombieType.digger,
    ZombieType.ladder,
    ZombieType.football,
    ZombieType.dancing
}
"""所有 ize 中出现的僵尸"""

plant_abbr_to_type: dict[str, PlantType | None] = {
    ".": None,
    "1": PlantType.pea_shooter,
    "h": PlantType.sunflower,
    "o": PlantType.wallnut,
    "t": PlantType.potato_mine,
    "b": PlantType.snow_pea,
    "z": PlantType.chomper,
    "2": PlantType.repeater,
    "p": PlantType.puffshroom,
    "d": PlantType.fumeshroom,
    "x": PlantType.scaredyshroom,
    "w": PlantType.squash,
    "3": PlantType.threepeater,
    "_": PlantType.spikeweed,
    "j": PlantType.torchwood,
    "l": PlantType.split_pea,
    "5": PlantType.starfruit,
    "c": PlantType.magnetshroom,
    "y": PlantType.kernelpult,
    "s": PlantType.umbrella_leaf
}
"""植物缩写到植物类型的字典"""

zombie_abbr_to_type: dict[str, ZombieType] = {
    "xg": ZombieType.imp,
    "lz": ZombieType.conehead,
    "cg": ZombieType.pole_vaulting,
    "tt": ZombieType.buckethead,
    "bj": ZombieType.bungee, "xt": ZombieType.bungee,
    "kg": ZombieType.digger,
    "tz": ZombieType.ladder, "ft": ZombieType.ladder,
    "gl": ZombieType.football,
    "ww": ZombieType.dancing, "mj": ZombieType.dancing
}
"""僵尸缩写到僵尸类型的字典"""


class Layout(Enum):
    HOTCHPOTCH = auto()
    """综合"""
    KERNELS = auto()
    """控制"""
    INSTANTS = auto()
    """即死"""
    PEAS = auto()
    """输出"""
    POTATOES = auto()
    """爆炸"""
    STARFRUITS = auto()
    """倾斜"""
    MAGNETS = auto()
    """穿刺"""
    SCAREDIES = auto()
    """回复"""



# -*- coding: utf_8 -*-
"""
子弹相关的枚举和类
"""
from enum import IntEnum
from typing import Self

from .obj_base import ObjNode, property_i32, property_f32, property_bool, property_int_enum, ObjId, obj_list
from ..basic import asm


class ProjectileType(IntEnum):
    """
    子弹类型
    """
    pea = 0x0
    snow_pea = 0x1
    cabbage = 0x2
    melon = 0x3
    puff = 0x4
    wintermelon = 0x5
    fire_pea = 0x6
    star = 0x7
    cactus = 0x8
    basketball = 0x9
    kernel = 0xA
    cob_cannon = 0xB
    butter = 0xC


class ProjectileMotionType(IntEnum):
    """
    子弹运动类型
    """
    straight = 0
    parabola = 1
    switch_way = 2
    puff = 5
    left_straight = 6
    starfruit = 7
    cattail = 9


class Projectile(ObjNode):
    """
    子弹对象
    """
    ITERATOR_FUNC_ADDRESS = 0x41C9B0

    OBJ_SIZE = 0x94

    int_x = property_i32(0x8, "图像整数 x 坐标")

    int_y = property_i32(0xc, "图像整数 y 坐标")

    col = property_i32(0x1c, "所在行数")

    x = property_f32(0x30, "浮点 x 坐标")

    y = property_f32(0x34, "浮点 y 坐标")

    dx = property_f32(0x3c, "x 速度")

    dy = property_f32(0x40, "y 速度")

    is_dead = property_bool(0x50, "是否死亡")

    type_ = property_int_enum(0x5c, ProjectileType, "子弹类型")

    motion_type = property_int_enum(0x58, ProjectileMotionType, "子弹运动类型")

    @property
    def target_zombie_id(self) -> ObjId:
        """香蒲子弹目标僵尸"""
        return ObjId(self.base_ptr + 0x88, self.controller)

    def die(self) -> None:
        """
        令自己死亡
        """
        code = f"""
            mov eax, {self.base_ptr}
            call {0x46EB20}  // Projectile::Die
            ret"""
        asm.run(code, self.controller)


class ProjectileList(obj_list(Projectile)):
    """
    子弹 DataArray
    """
    def free_all(self) -> Self:
        code = f"""
                push edi   
                push esi
                mov eax, [0x6a9ec0]
                mov edi, [eax + 0x768]
                mov esi, {self.controller.result_address}
                xor edx, edx
                mov [esi], edx
                LIterate:
                    mov {Projectile.ITERATOR_P_BOARD_REG}, edi
                    call {Projectile.ITERATOR_FUNC_ADDRESS}  // Board::IterateProjectile
                    test al, al
                    jz LFreeAll
                    mov eax, [esi]
                    call {0x46EB20}  // Projectile::Die
                    jmp LIterate
                    
                LFreeAll:
                    mov edi, {self.base_ptr}
                    call {0x41e600}  // DataArray<Zombie>::DataArrayFreeAll
                    pop esi
                    pop edi
                    ret"""
        asm.run(code, self.controller)
        return self



# -*- coding: utf_8 -*-
"""
iztools 全场测试功能模拟
"""
import time
from collections.abc import Callable
from msvcrt import kbhit, getwch
from random import randint
from typing import TypeAlias, Self, overload, NamedTuple, SupportsIndex
import warnings

from .consts import plant_abbr_to_type, zombie_abbr_to_type
from .operations import enter_ize
from .plant_modifier import randomize_generate_cd
from ..basic.gridstr import parse_grid_str, GridStr
from ..basic.inject import ConnectedContext
from ..flow.flow import FlowFactory, TickRunnerResult, FlowManager, DEFAULT_PRIORITY
from ..flow.utils import until
from ..rp_extend import Controller, HookPosition, RpBaseException
from ..structs.game_board import GameBoard, get_board
from ..structs.griditem import Griditem
from ..structs.plant import PlantType, Plant
from ..structs.zombie import ZombieType, Zombie


class PlaceZombieOp(NamedTuple):
    """
    描述僵尸放置操作的对象

    Attributes:
        type_: 要放置的僵尸类型
        time: 放置的时间
        row: 放置的行, 从0开始
        col: 放置的列, 从0开始
    """
    type_: ZombieType
    time: int
    row: int
    col: int


PlantTypeList: TypeAlias = list[list[PlantType | None]]


def parse_plant_type_list(plant_type_str: str) -> tuple[PlantTypeList, PlantTypeList]:
    """
    根据 iztools 植物字符串生成植物列表

    使用+号表示延迟一轮种植, 即"调控栈位".

    Args:
        plant_type_str: 与 izt 要求相同的植物列表字符串.
    Returns:
        两个5 * 5列表, 分别表示第一轮, 第二轮种植的植物. 空白值由 None 填充
    Raises:
        ValueError: plant_type_string 格式错误时抛出
    """
    first_list: list[list[PlantType | None]] = [[None] * 5 for _ in range(5)]
    second_list: list[list[PlantType | None]] = [[None] * 5 for _ in range(5)]
    lines = plant_type_str.strip().splitlines(False)
    if (t := len(lines)) != 5:
        raise ValueError(f"plant_type_string must have 5 lines, instead of {t} lines")
    for row, line in enumerate(lines):
        line = line.strip()
        plus_plant_indices = [i - 1 for i, char in enumerate(line) if char == '+']
        if not plus_plant_indices:
            if (t := len(line)) != 5:
                raise ValueError(f"line {row} must have 5 plants, instead of {t} plants")
            first_list[row] = [plant_abbr_to_type[abbr] for abbr in line]
        else:
            if plus_plant_indices[0] == -1:
                raise ValueError(f"line {row} can't start with +")
            col = 0
            for (i, char) in enumerate(line):
                if char == "+":
                    continue
                if i in plus_plant_indices:
                    second_list[row][col] = plant_abbr_to_type[char]
                else:
                    first_list[row][col] = plant_abbr_to_type[char]
                col += 1
            if col != 5:
                raise ValueError(f"line {row} must have 5 plants, instead of {t} plants")
    return first_list, second_list


def parse_target_list(target_str: str) -> tuple[list[tuple[int, int]], list[int]]:
    """
    根据 iztools 目标字符串生成目标列表

    Args:
        target_str: 与 izt 要求相同的目标字符串
    Returns:
        两个列表, 分别表示目标植物和目标脑子的位置
    """
    poses = [parse_grid_str(pos) for pos in target_str.strip().split()]
    return [pos for pos in poses if pos[1] != -1], [pos[0] for pos in poses if pos[1] == -1]


def parse_zombie_place_list(place_zombie_str: str) -> list[PlaceZombieOp]:
    """
    根据 iztools 僵尸放置字符串生成僵尸放置列表

    Args:
        place_zombie_str: 与 izt 要求相同的僵尸放置字符串.
    Returns:
        一个列表, 返回所有僵尸操作, 用 PlaceZombieOp 表示
    Raises:
        ValueError: place_zombie_string 格式错误时抛出
    """
    lines = place_zombie_str.strip().splitlines(False)
    if (t := len(lines)) != 3:
        raise ValueError(f"place_zombie_string must have 3 lines, instead of {t} lines")
    types = [zombie_abbr_to_type[abbr] for abbr in lines[0].strip().split()]
    times = [int(time_) for time_ in lines[1].strip().split()]
    rows, cols = zip(*(parse_grid_str(pos) for pos in lines[2].strip().split()))  # zip(*iterable)转置
    if not (len(types) == len(times) == len(rows) == len(cols)):
        raise ValueError("length of types, times, rows and cols must be equal")
    return [PlaceZombieOp(*op) for op in zip(types, times, rows, cols)]


_Id: TypeAlias = tuple[int, int]


class _IzGround:
    def __init__(self, origin_plant_ids: list[list[_Id]], origin_brain_ids: list[_Id], izt: "IzTest") -> None:
        self.origin_plant_ids: list[list[_Id]] = origin_plant_ids
        self.origin_brain_ids: list[_Id] = origin_brain_ids
        self.zombie_ids: list[_Id] = []
        self.izt: "IzTest" = izt

    @overload
    def __getitem__(self, item: tuple[int, int]) -> Plant | Griditem | None:
        """
        通过(row, col)获得测试开始时对应位置的植物或脑子.

        Args:
            item: (row, col)元组
        Returns:
            对象不存在 or 已死亡返回None, 否则返回该植物/脑子.
        Examples:
            >>> ground: _IzGround = ...
            >>> plant = ground[0, 0]  # 获得1-1位置的植物
            >>> brain = ground[4, -1]  # 获得第5行的脑子
        """

    @overload
    def __getitem__(self, item: GridStr) -> Plant | Griditem | None:
        """
        通过 GridStr 获得测试开始时对应位置的植物或脑子.

        Args:
            item: GridStr 位置
        Returns:
            对象不存在 or 已死亡返回None, 否则返回该植物/脑子.
        Examples:
            >>> ground: _IzGround = ...
            >>> plant = ground["1-1"]  # 获得1-1位置的植物
            >>> brain = ground["5-0"]  # 获得第5行的脑子
        """

    def __getitem__(self, item):
        match item:
            case (row, -1):
                t = self.izt.game_board.griditem_list.find(*self.origin_brain_ids[row])
                return None if t is None or t.is_dead else t
            case (row, col):
                t = self.izt.game_board.plant_list.find(*self.origin_plant_ids[row][col])
                return None if t is None or t.is_dead else t
            case grid:
                return self.__getitem__(parse_grid_str(grid))

    def zombie(self, i: SupportsIndex) -> Zombie | None:
        """
        获得写在 IzTest init_str 上的第 i 个僵尸

        Args:
            i: 索引, 支持负数, 如-1表示"此时 init_str 上写的, 最近放置的僵尸"
        Returns:
            不存在 or 已死亡返回 None, 否则返回僵尸
        """
        try:
            id_ = self.zombie_ids[i]
        except IndexError:
            return None
        t = self.izt.game_board.zombie_list.find(*id_)
        return None if t is None or t.is_dead else t


class IzTest:
    """
    模拟iztools全场测试.

    Attributes:
        plant_type_lists: 两个5 * 5列表, 分别表示第一轮, 第二轮种植的植物. 空白值由 None 填充.
        place_zombie_list: 一个列表, 表示所有僵尸操作, 用 PlaceZombieOp 表示.
        repeat_time: 重复次数.
        mj_init_phase: mj 初始相位. None 表示随机
        target_plants_pos: 目标植物的位置列表. 元素为(row, col)
        target_brains_pos: 目标脑子的位置列表. 元素为 row
        controller: 测试使用的 Controller 对象.
        flow_factory: 生成测试逻辑的 FlowFactory 对象.
        reset_generate_cd: 是否重置植物的 generate_cd , 即, iztools"开启攻击间隔处理" is True
        enable_default_check_end: 是否启用默认的判断输赢功能, 即, 需要手动设置判断时为 False
        start_check_end_time: 开始判断一次测试是否输赢的时间, 默认为放下最后一个僵尸的时间.
        end_callback: 一次测试结束时的回调函数, 参数为是否成功 bool.
        check_tests_end_callback: 判断是否结束测试的回调函数. 默认为 None, 表示重复 repeat_time 次测试.
            参数为(当前测试次数, 成功次数), 返回 None 表示不结束, 返回 float 表示结果概率.
        ground: 用于获取原始植物和脑子的对象, 仅在测试中调用有效.

    """

    def __init__(self, controller: Controller, reset_generate_cd: bool = True):
        """
        构造 IzTest 对象

        如此构造的对象不能直接使用! 大部分情况下需要调用 init_by_str 初始化.

        Args:
            controller: 测试使用的 Controller 对象.
            reset_generate_cd: 攻击间隔处理 in iztools
        """
        self.plant_type_lists: tuple[PlantTypeList, PlantTypeList] = ([], [])
        self.place_zombie_list: list[PlaceZombieOp] = []
        self.repeat_time: int = 0
        self.mj_init_phase: int | None = None
        self.target_plants_pos: list[tuple[int, int]] = []
        self.target_brains_pos: list[int] = []
        self.controller: Controller = controller
        self.flow_factory: FlowFactory = FlowFactory()
        self.reset_generate_cd: bool = reset_generate_cd
        self.enable_default_check_end: bool = False
        self.start_check_end_time: int = 0
        self.end_callback: Callable[[bool], None] = lambda _: None
        self.check_tests_end_callback: Callable[[int, int], float | None] | None = None
        self.ground: _IzGround | None = None

        # 运行时候会时刻改变的量. 不建议修改
        self._target_plant_ids: list[tuple[int, int]] = []  # 所有目标脑子
        self._target_brain_ids: list[tuple[int, int]] = []  # 所有目标植物
        self._last_test_ended: bool = False  # 用于判断是否结束一次测试
        self._success_count: int = 0  # 成功次数
        self._test_time: int = 0  # 测试次数

        self._flow_factory_set: bool = False  # 用于判断是否设置了flow_factory

    @property
    def game_board(self) -> GameBoard:
        """游戏 GameBoard 对象"""
        return get_board(self.controller)

    def init_by_str(self, iztools_str: str) -> Self:
        """
        通过 iztools 字符串初始化 iztest 对象

        与 iztools 的输入格式不完全相同:
            - 允许首尾空行以及每行首尾空格.
            - 支持“测试次数”输入-1表示自定义结束行为:
                结束行为默认为测试无限次, 可以通过 self.check_tests_end() 设置何时结束.
            - 支持第二行空行表示无目标: 若此行为空, 则不启用内置的判断输赢功能.
            - 支持不输入8 9 10行表示不放置僵尸: 若此三行为空, 则不启用内置的判断输赢功能.
            - (暂且)不支持通过书写顺序调整僵尸编号, 可以通过 ObjList 相关接口调整.

        Args:
            iztools_str: iztools输入字符串
        Returns:
            self
        Raises:
            ValueError: 输入字符串格式错误时抛出
        Examples:
            >>> ctler: Controller = ...
            >>> iz_test = IzTest(ctler).init_by_str('''
            ...     1000 -1
            ...     3-0 4-0 5-0 3-3
            ...     .....
            ...     .....
            ...     bs3_c
            ...     b2ljh
            ...     blyl_
            ...     cg   cg   xg   ww
            ...     0    1    300  700
            ...     4-6  4-6  4-6  4-6''')
            如上为iztools默认例子的输入方式.
        """
        lines = iztools_str.strip().splitlines(False)

        if len(lines) == 7:
            self.place_zombie_list = []
        elif len(lines) == 10:
            self.place_zombie_list = parse_zombie_place_list('\n'.join(lines[7:10]))
            self.start_check_end_time = max(op.time for op in self.place_zombie_list)
        else:
            raise ValueError(f"iztools_str must have 7 or 10 lines, not {len(lines)} lines")

        repeat_time, mj_init_phase = map(int, lines[0].strip().split())
        if mj_init_phase < -1 or mj_init_phase >= 460:
            raise ValueError(f"mj_init_phase must be in [-1, 459], not {mj_init_phase}")
        self.mj_init_phase = mj_init_phase if mj_init_phase != -1 else None
        if repeat_time < -1 or repeat_time == 0:
            raise ValueError(f"repeat_time must be positive or -1, not {repeat_time}")
        if repeat_time != -1:
            self.repeat_time = repeat_time
        else:
            self.check_tests_end_callback = lambda _, __: None

        self.target_plants_pos, self.target_brains_pos = parse_target_list(lines[1])
        if self.target_plants_pos or self.target_brains_pos:
            self.enable_default_check_end = True

        self.plant_type_lists = parse_plant_type_list('\n'.join(lines[2:7]))
        for target_pos in self.target_plants_pos:
            if (self.plant_type_lists[0][target_pos[0]][target_pos[1]] is None and
                    self.plant_type_lists[1][target_pos[0]][target_pos[1]] is None):
                raise ValueError(f"target plant at {target_pos} is None")
        return self

    def on_game_end(self) -> Callable[[Callable[[bool], None]], Callable[[bool], None]]:
        """
        装饰器, 设置结束时的回调函数

        回调的 bool 参数为本次测试是否成功.

        Returns:
            添加用装饰器
        """

        def _decorator(func):
            self.end_callback = func
            return func

        return _decorator

    def end(self, succeeded: bool) -> TickRunnerResult:
        """
        返回本函数, 表示本次测试结束

        Args:
            succeeded: 成功则传入 True, 失败传入 False
        Returns:
            TickRunnerResult.BREAK_RUN
        """
        self._last_test_ended = True
        self.end_callback(succeeded)
        if succeeded:
            self._success_count += 1
        self._test_time += 1

        self._target_plant_ids = []
        self._target_brain_ids = []

        return TickRunnerResult.BREAK_DONE

    def check_end(self) -> TickRunnerResult | None:
        """
        默认的判断是否结束测试的函数

        Returns:
            如果结束则返回 TickRunnerResult.BREAK_RUN, 否则返回 None
        """
        board = self.game_board
        if (all(board.griditem_list.find(*brain) is None for brain in self._target_brain_ids) and
                all(board.plant_list.find(*plant) is None for plant in self._target_plant_ids)):
            return self.end(True)
        if board.zombie_list.obj_num == 0:
            return self.end(False)

    def check_tests_end(self) \
            -> Callable[[Callable[[int, int], float | None]], Callable[[int, int], float | None]]:
        """
        装饰器, 设置判断是否结束测试的回调函数

        Returns:
            添加用装饰器
        """

        def _decorator(func):
            self.check_tests_end_callback = func
            return func

        return _decorator

    def set_flow_factory(self,
                         place_priority: int = DEFAULT_PRIORITY + 10,
                         check_end_priority: int = DEFAULT_PRIORITY - 10) -> Self:
        """
        设置 flow_factory

        Args:
            place_priority: 初始化及放置僵尸 tick runner 的优先级, 默认为 default + 10
            check_end_priority: 判断输赢 tick runner 的优先级, 默认为 default - 10
        Returns:
            self
        Raises:
            RpBaseException: 已经设置过 flow_factory 时抛出
        """
        if self._flow_factory_set:
            raise RpBaseException("cannot set flow factory twice!")
        self._flow_factory_set = True

        @self.flow_factory.connect(until(0), only_once=True, priority=place_priority)
        def _init(_):
            # 清掉所有_ObjList的栈
            origin_plant_ids: list[list[_Id]] = [[None] * 5 for _ in range(5)]  # type: ignore
            origin_brain_ids: list[_Id] = [None] * 5  # type: ignore
            board = self.game_board
            board.plant_list.free_all().reset_stack()
            board.zombie_list.free_all().reset_stack()
            board.projectile_list.free_all().reset_stack()
            board.griditem_list.free_all().reset_stack()
            board.mj_clock = randint(0, 459) if self.mj_init_phase is None else self.mj_init_phase

            for plant_list in self.plant_type_lists:
                for row, line in enumerate(plant_list):
                    for col, type_ in enumerate(line):
                        if type_ is None:
                            continue
                        plant = board.iz_new_plant(row, col, type_)
                        # assert plant is not None
                        origin_plant_ids[row][col] = plant.id.tpl()
                        if self.reset_generate_cd:
                            randomize_generate_cd(plant)
                        if (row, col) in self.target_plants_pos:
                            self._target_plant_ids.append(plant.id.tpl())

            for i in range(5):
                brain = self.game_board.new_iz_brain(i)
                origin_brain_ids[i] = brain.id.tpl()
                if i in self.target_brains_pos:
                    self._target_brain_ids.append(brain.id.tpl())

            self.ground = _IzGround(origin_plant_ids, origin_brain_ids, self)

        for op in self.place_zombie_list:
            @self.flow_factory.add_tick_runner(place_priority)
            def _place_zombie(fm: FlowManager, _op=op):
                if fm.time == _op.time:
                    t = self.game_board.iz_place_zombie(_op.row, _op.col, _op.type_)
                    self.ground.zombie_ids.append(t.id.tpl())
                    return TickRunnerResult.DONE

        if self.enable_default_check_end:
            @self.flow_factory.add_tick_runner(check_end_priority)
            def _check_end(fm: FlowManager):
                if fm.time >= self.start_check_end_time:
                    return self.check_end()
        return self

    def start_test(self, jump_frame: bool = False,
                   speed_rate: float = 1.0,
                   print_interval: int = 10,
                   control_speed_key: str = '\x12'  # deprecated
                   ) -> tuple[float, float]:
        """
        开始测试

        Args:
            jump_frame: True则开启跳帧测试.
            speed_rate: 速度倍率. 仅当 jump_frame = False 时有效.
            print_interval: 每隔 print_interval 次测试打印一次结果. 输入0时代表不打印
            control_speed_key: [[deprecated]] 非跳帧时切换原速/倍速的按键. 默认值为 Ctrl+R
        Returns:
            (测试概率, 使用时间)元组
        """
        if control_speed_key != '\x12':
            warnings.warn("deprecated param `control_speed_key`", DeprecationWarning)
        if self.controller.read_i32(0x6a9ec0, 0x7f8) != 70:  # gLawnApp->mGameMode == ize
            enter_ize(self.controller)
        start_time = time.time()
        last_time = start_time
        ctler = self.controller
        ctler.open_hook(HookPosition.CHALLENGE_I_ZOMBIE_SCORE_BRAIN)
        if not self._flow_factory_set:
            self.set_flow_factory()
        with ConnectedContext(ctler) as ctler:
            if jump_frame:
                ctler.start_jump_frame()
            else:
                frame_duration = 1 if (fd := round(10 / speed_rate)) == 0 else fd
                self.game_board.frame_duration = frame_duration
            ctler.skip_frames()

            def _one_test():
                nonlocal last_time
                _flow_manager = self.flow_factory.build_manager()
                ctler.skip_frames()
                while not self._last_test_ended:
                    _flow_manager.run()
                    if not jump_frame and kbhit() and getwch() == control_speed_key:
                        self.game_board.frame_duration = 10 \
                            if self.game_board.frame_duration != 10 else frame_duration
                    # print(_flow_manager.time)
                    ctler.skip_frames()
                self._last_test_ended = False
                _flow_manager.end()
                if print_interval and self._test_time % print_interval == 0:
                    print(f"ended {self._test_time} of {self.repeat_time}, "
                          f"success rate: {self._success_count / self._test_time:.2%}, "
                          f"using time: {(t := time.time()) - last_time:.2f}s.")
                    last_time = t

            if _callback := self.check_tests_end_callback:
                _one_test()  # no do while!
                while (result := _callback(self._test_time, self._success_count)) is None:
                    _one_test()
            else:
                for _ in range(self.repeat_time):
                    _one_test()
                result = self._success_count / self.repeat_time

            if jump_frame:
                ctler.end_jump_frame()
            else:
                self.game_board.frame_duration = 10
            ctler.close_hook(HookPosition.CHALLENGE_I_ZOMBIE_SCORE_BRAIN)
        return result, (time.time() - start_time)



# -*- coding: utf_8 -*-
"""
异常类.
"""
from ..rp_extend import RpBaseException


class PvzStatusError(RpBaseException):
    """
    pvz 状态异常

    当 pvz 游戏状态不满足要求时抛出.
    """


class AsmError(RpBaseException):  # 作为 keystone error 包装使用, 万一哪天换汇编器了
    """
    汇编码编译异常

    当汇编码执行出现问题时抛出.
    """



# -*- coding: utf_8 -*-
"""
mj相关操控
"""
from collections.abc import Callable
from contextlib import AbstractContextManager
from enum import Enum
from typing import SupportsIndex, Literal, TypeAlias, Self, overload

from .iztest import IzTest
from ..flow.flow import FlowManager, TickRunnerResult, CondFunc, DEFAULT_PRIORITY
from ..flow.utils import AwaitableCondFunc
from ..rp_extend import ControllerError
from ..structs.game_board import GameBoard, get_board
from ..structs.zombie import Zombie, ZombieStatus

BackupPos: TypeAlias = Literal["w", "s", "a", "d"]
"""伴舞相对位置, w 上, s 下, a 前, d 后"""


@overload
def partner(mj: Zombie, item: SupportsIndex | BackupPos) -> Zombie | None:
    """
    获得 mj 的伴舞位置

    Args:
        mj: 目标 mj
        item: 伴舞位置, 支持下标和字符串; 使用下表时按照伴舞位置的顺序, 0上, 1下, 2前, 3后
    Returns:
        伴舞位置的 Zombie 对象, 如果没有则返回 None
    Raises:
        ValueError: 如果 item 不是有效位置
    """


@overload
def partner(mj: Zombie, item: str) -> tuple[Zombie | None, ...]:
    """
    获得 mj 的伴舞位置

    Args:
        mj: 目标 mj
        item: 伴舞位置字符串, 按顺序返回

    Returns:
        一个元组, 按 item 中顺序返回伴舞位置的 Zombie 对象, 如果没有则返回 None
    Examples:
        >>> partner(mj, "ws") == tuple(partner(mj, "w"), partner(mj, "s"))
        True
        >>> front, back = partner(mj, "ad")  # 获得前后伴舞
    """


def partner(mj: Zombie, item):
    zlist = get_board(mj.controller).zombie_list
    partners = [zlist.find(id_) for id_ in mj.partner_ids]
    if isinstance(item, SupportsIndex):
        return partners[item]
    if len(item) != 1:
        return tuple(partner(mj, i) for i in item)
    match item:
        case "w":
            return partners[0]
        case "s":
            return partners[1]
        case "a":
            return partners[2]
        case "d":
            return partners[3]
        case _:
            raise ValueError(f"{item} is not a valid partner pos")


def get_clock(board: GameBoard | None = None) -> int:
    """
    获取当前时钟

    Args:
        board: 默认为使用 get_board()
    Returns:
        当前时钟
    """
    if board is None:
        board = get_board()
    if (clock := board.mj_clock) >= 0:
        return clock % 460
    return -(-clock % 460)  # c/c++ %


def get_dancing_status(clock: int) -> ZombieStatus:
    """
    获取时钟对应的 dancing 状态

    Args:
        clock: 时钟时间
    Returns:
        当前时钟对应的僵尸状态
    """
    match clock:
        case t if t <= 11:
            return ZombieStatus.dancing_walking
        case 12:
            return ZombieStatus.dancing_armrise1  # 尝试召唤伙伴相位
        case 13 | 14 | 15:
            return ZombieStatus.dancing_armrise3
        case 16 | 17 | 18:
            return ZombieStatus.dancing_armrise2
        case 19 | 20 | 21:
            return ZombieStatus.dancing_armrise5
        case _:
            return ZombieStatus.dancing_armrise4


class DancingPhase(Enum):
    """
    mj相位, self.value表示该相位下的**一个**时钟值
    """
    TRYING_CALLING_PARTNER = 240  # 12 * 20
    DANCING = 260  # 13 * 20
    MOVING = 0


DancingPhaseLiteral: TypeAlias = DancingPhase | Literal["summon", "dance", "move", "s", "d", "m"]
"""mj 相位字面量"""


def to_dancing_phase(literal: DancingPhaseLiteral) -> DancingPhase:
    """
    将字面量转换为DancingPhase

    Args:
        literal: 字面量
    Returns:
        转换后的 DancingPhase 对象
    Raises
        ValueError: 如果 literal 不是有效的相位字面量
    """
    match literal:
        case "summon" | "s" | DancingPhase.TRYING_CALLING_PARTNER:
            return DancingPhase.TRYING_CALLING_PARTNER
        case "dance" | "d" | DancingPhase.DANCING:
            return DancingPhase.DANCING
        case "move" | "m" | DancingPhase.MOVING:
            return DancingPhase.MOVING
        case _:
            raise ValueError(f"{literal} is not a dancing phase literal")


def dancing_status_to_phase(status: ZombieStatus) -> DancingPhase:
    """
    将僵尸状态转换为 DancingPhase

    Args:
        status: 僵尸状态
    Returns:
        转换后的 DancingPhase 对象
    Raises:
        ValueError: 如果 status 不是 mj 特有的状态
    """
    match status:
        case ZombieStatus.dancing_walking:
            return DancingPhase.MOVING
        case ZombieStatus.dancing_armrise1:
            return DancingPhase.TRYING_CALLING_PARTNER
        case ZombieStatus.dancing_armrise2 | ZombieStatus.dancing_armrise3 | (
             ZombieStatus.dancing_armrise4) | ZombieStatus.dancing_armrise5:
            return DancingPhase.DANCING
        case _:
            raise ValueError(f"{status} is not a dancing phase")


def get_dancing_phase(board: GameBoard | None = None) -> DancingPhase:
    """
    获取当前时钟对应的 DancingPhase

    Args:
        board: 游戏板对象, 默认为使用get_board()
    Returns:
        当前时钟对应的 DancingPhase
    """
    return dancing_status_to_phase(get_dancing_status(get_clock(board)))


class _DmTr(Callable):
    def __init__(self, iz_test: IzTest):
        self.iz_test = iz_test
        self.current_phase: DancingPhase | None = None  # None表示不控制
        self.next_phase: DancingPhase | None = None  # None表示没有"下一个"状态
        self.cond_to_next: CondFunc = lambda _: True

    def keep_phase(self, _: FlowManager) -> None:
        cp = self.current_phase
        board = get_board(self.iz_test.controller)
        if cp.value != get_clock(board):
            board.mj_clock = cp.value

    def switch_to_next_phase(self, fm: FlowManager) -> None:
        if not self.cond_to_next(fm):
            return
        self.current_phase = self.next_phase
        self.next_phase = None
        self.cond_to_next = lambda _: True

    def __call__(self, fm: FlowManager) -> TickRunnerResult | None:
        if self.next_phase is not None:
            self.switch_to_next_phase(fm)
        if self.current_phase is not None:
            self.keep_phase(fm)
        return None

    @property
    def waiting_next_phase(self) -> bool:
        return self.next_phase is not None

    @property
    def is_controlling(self) -> bool:
        return self.current_phase is not None

    def start_controlling(self, phase: DancingPhase) -> None:
        self.current_phase = phase

    def stop_controlling(self, phase: DancingPhase) -> None:
        get_board(self.iz_test.controller).mj_clock = phase.value
        self.current_phase = None
        self.next_phase = None


class DancingManipulator(AbstractContextManager):
    """
    mj 相位控制器, 即, "女仆秘籍".

    与 AvZ 不同, DancingManipulator 目前依托于一个T ickRunner 控制mj时钟工作.
    构造函数为 protected interface.

    Attributes:
        start_phase: with 语句开始时的相位
        end_phase: with 语句结束时的相位
    """
    def __init__(self, tr: _DmTr,
                 start_phase: DancingPhaseLiteral,
                 end_phase: DancingPhaseLiteral):  # protected interface
        self.start_phase: DancingPhase = to_dancing_phase(start_phase)
        self.end_phase: DancingPhase = to_dancing_phase(end_phase)
        self._tr: _DmTr = tr

    def next_phase(self, phase: DancingPhaseLiteral,
                   condition: CondFunc = lambda _: True) -> None:
        """
        设置下一个相位

        Args:
            phase: 下一个相位
            condition: 到下一个相位的条件, 默认为恒真, 即直接进入下一个相位
        """
        self._tr.cond_to_next = condition
        self._tr.next_phase = to_dancing_phase(phase)

    async def until_next_phase(self, phase: DancingPhaseLiteral, condition: CondFunc):
        """
        等待到下一个相位, 用于 flow 内部

        使用中等效于:
        >>> async def flow(_):
        ...     await until(condition)  # noqa
        ...     self.next_phase(phase)

        Args:
            phase: 下一个相位
            condition: 到下一个相位的条件
        """
        await AwaitableCondFunc(condition)
        self.next_phase(phase)

    def stop(self, end_phase: DancingPhaseLiteral) -> None:
        """
        停止控制

        Args:
            end_phase: 停止控制时的相位
        """
        if self._tr.is_controlling:
            self._tr.stop_controlling(to_dancing_phase(end_phase))

    def start(self, first_phase: DancingPhaseLiteral) -> Self:
        """
        开始控制

        Args:
            first_phase: 开始控制时的相位
        Returns:
            self
        """
        if not self._tr.is_controlling:
            self._tr.start_controlling(to_dancing_phase(first_phase))
        return self

    def __enter__(self) -> Self:
        return self.start(self.start_phase)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ControllerError:
            return
        self.stop(self.end_phase)


def get_dancing_manipulator(iz_test: IzTest,
                            start_phase: DancingPhaseLiteral = DancingPhase.MOVING,
                            end_phase: DancingPhaseLiteral = DancingPhase.MOVING,
                            priority: int = DEFAULT_PRIORITY + 1) -> DancingManipulator:
    """
    获取一个 DancingManipulator

    Args:
        iz_test: 和 DancingManipulator 对应的 IzTest 对象
        start_phase: 开始时的相位
        end_phase: 结束时的相位
        priority: DancingManipulator 依托的 TickRunner 优先级.
            默认为 default + 1, 以确保默认时 DancingManipulator 在本帧生效
    Returns:
        构造的 DancingManipulator 对象
    """
    dm_tr = _DmTr(iz_test)
    iz_test.flow_factory.add_tick_runner(priority)(dm_tr)
    return DancingManipulator(dm_tr, start_phase, end_phase)



# -*- coding: utf_8 -*-
"""
ize 漏阳光相关函数
"""
from ..basic.gridstr import GridStr, get_grid_str
from ..rp_extend import RpBaseException
from ..structs.game_board import GameBoard, get_board
from ..structs.plant import Plant, PlantType

SUNFLOWER_HPS_ON_DROPPING_SUN: tuple[int, int, int, int, int, int, int, int] = \
    tuple(276 - 40 * i for i in range(7)) + (0,)  # type: ignore
"""向日葵掉落阳光时的血量"""


def get_sunflower_remaining_sun(sunflower: Plant | None) -> int:
    """
    获取向日葵剩余阳光

    Args:
        sunflower: 目标向日葵
    Returns:
        向日葵剩余阳光值, 输入 None 时返回0
    Raises:
        RpBaseException: 出现未知的行为时
    """
    if sunflower is None or sunflower.is_dead:
        return 0
    hp = sunflower.hp
    intervals = (300,) + SUNFLOWER_HPS_ON_DROPPING_SUN
    lo, hi = 0, len(intervals) - 1
    while lo <= hi:
        mid = (lo + hi) // 2
        if intervals[mid] >= hp > intervals[mid + 1]:
            return (8 - mid) * 25
        if hp > intervals[mid]:
            hi = mid - 1
        else:
            lo = mid + 1
    raise RpBaseException("unexpected behaviour!")


def get_all_remaining_suns(board: GameBoard | None = None) -> dict[GridStr, int]:
    """
    获取所有向日葵剩余阳光

    Args:
        board: 要获取的 board. 为 None 时使用 get_board()
    Returns:
        一个字典, 形式为{向日葵所在的 x-y: 剩余阳光值}
    """
    if board is None:
        board = get_board()
    return {get_grid_str(sf.row, sf.col): get_sunflower_remaining_sun(sf)
            for sf in ~board.plant_list if sf.type_ is PlantType.sunflower}



# -*- coding: utf_8 -*-
"""
场地物品相关的枚举和类
"""
from enum import IntEnum
from typing import Self

from .obj_base import ObjNode, property_int_enum, property_i32, property_bool, property_f32, obj_list
from ..basic import asm


class GriditemType(IntEnum):
    """
    场地物品类型
    """
    none = 0
    grave = 0x1
    crater = 0x2
    ladder = 0x3
    brain_aq = 0x6
    vase = 0x7
    rake = 0xb
    brain = 0xc


class Griditem(ObjNode):
    """
    场地物品. 包括脑子, 梯子等
    """
    OBJ_SIZE = 0xEC

    ITERATOR_FUNC_ADDRESS = 0x41CAD0

    type_ = property_int_enum(0x8, GriditemType, "场地物品类型")

    row = property_i32(0x14, "所在行")

    col = property_i32(0x10, "所在列")

    brain_hp = property_i32(0x18, """
        脑子血量, 墓碑冒出的量, 弹坑消失倒计时, 钉钯消失倒计时
        
        对于 ize 中脑子, 初始为70, 每次被啃时-= 1(区别于植物血量 -= 4)
        """)

    layer = property_i32(0x1c, "图层")

    is_dead = property_bool(0x20, "是否死亡")

    x = property_f32(0x24, "x 坐标")

    y = property_f32(0x28, "y 坐标")

    def __str__(self) -> str:
        if not self.is_dead:
            return f"#{self.id.index} {self.type_.name} at {self.row + 1}-{self.col + 1}"
        return "dead griditem"

    def die(self) -> None:
        """
        令自己死亡
        """
        code = f"""
            push esi
            mov esi, {self.base_ptr}
            call {0x44D000}  // Griditem::GriditemDie
            ret"""
        asm.run(code, self.controller)


class GriditemList(obj_list(Griditem)):
    """
    场地物品 DataArray
    """
    def alloc_item(self) -> Griditem:
        """
        从内存数组中申请新的 griditem 对象

        Returns:
            申请出的 Griditem 对象
        """
        code = f"""
            push esi
            mov esi, {self.base_ptr}
            call {0x41E1C0}  // DataArray<GridItem>::DataArrayAlloc
            mov [{self.controller.result_address}], eax
            pop esi
            ret"""
        asm.run(code, self.controller)
        return Griditem(self.controller.result_u32, self.controller)

    def free_all(self) -> Self:
        code = f"""
                push edi
                push esi
                mov eax, [0x6a9ec0]
                mov edi, [eax + 0x768]
                mov esi, {self.controller.result_address}
                xor edx, edx
                mov [esi], edx
                LIterate:
                    mov {Griditem.ITERATOR_P_BOARD_REG}, edi
                    call {Griditem.ITERATOR_FUNC_ADDRESS}  // Board::IterateGriditem
                    test al, al
                    jz LFreeAll
                    mov esi, [esi]
                    call {0x44D000}  // Griditem::GriditemDie
                    mov esi, {self.controller.result_address}
                    jmp LIterate
                    
                LFreeAll:
                    mov eax, {self.base_ptr}
                    call {0x41E7D0}  // DataArray<Griditem>::DataArrayFreeAll
                    pop esi
                    pop edi
                    ret"""
        asm.run(code, self.controller)
        return self



# -*- coding: utf_8 -*-
"""
单刀赴会 脚本示例; 女仆脚本演示
"""
from ..flow.utils import until
from ..iztest.dancing import get_dancing_manipulator, partner
from ..iztest.iztest import IzTest
from ..rp_extend import Controller
from ..structs.zombie import ZombieStatus


def dancing_example(ctler: Controller, jump_frame=False):
    iz_test = IzTest(ctler).init_by_str("""
        1000 -1
        1-0 2-5 2-0
        ppo5s
        1ljys
        ...3.
        .....
        .....
        mj
        0
        1-9
        """)

    dm = get_dancing_manipulator(iz_test, "move")

    @iz_test.flow_factory.add_tick_runner()
    def extra_check_end(_):
        if iz_test.ground.zombie(0) is None:  # mj死亡直接判负, 以免测试卡死
            return iz_test.end(False)

    @iz_test.flow_factory.add_flow()
    async def flow(_):
        board = iz_test.game_board
        mj = iz_test.ground.zombie(0)
        await until(lambda _: board.zombie_list.obj_num >= 4)
        first_front, first_back = partner(mj, "ad")

        with dm:  # 等召唤出来开始控制
            await dm.until_next_phase("summon", lambda _: mj.x < 350)  # 走到5列开始站着召唤
            await until(lambda _: first_front.is_dead)  # 等前伴舞死后开始运后伴舞
            while True:
                await dm.until_next_phase(
                    "move", lambda _:
                    mj.status is ZombieStatus.dancing_summoning)  # 运后伴舞: 等mj召唤的时候相位切前进
                await dm.until_next_phase(
                    "summon", lambda _:
                    mj.status is not ZombieStatus.dancing_summoning  # 运后伴舞: 等mj召唤结束切回试图召唤
                    or first_back.int_x < mj.int_x - 5)  # 特判结束条件: 前伴舞走到mj前面一点的时候
                if first_back.int_x < mj.int_x - 5:
                    dm.next_phase("dance")  # 前伴舞走到mj前面一点的时候切换到跳舞, 省着走到太前面导致后续暴毙
                    await dm.until_next_phase(
                        "move", lambda _:
                        mj.status is not ZombieStatus.dancing_summoning)  # 等mj召唤结束切到走路相位
                    break
            await dm.until_next_phase(
                "summon", lambda _:
                mj.x < 205 or (not first_back.is_dead and first_back.is_eating))  # 到三列了停下来运后伴舞
            # print(first_back, first_back.int_x, mj.int_x)

            while True:
                await dm.until_next_phase(
                    "move", lambda _:
                    mj.status is ZombieStatus.dancing_summoning)
                if iz_test.ground["1-3"] is None:  # 运后伴舞, 这里要求少点等坚果死了往前走就行
                    dm.next_phase("move")
                    break
                await dm.until_next_phase(
                    "summon", lambda _:
                    mj.status is not ZombieStatus.dancing_summoning)
            await dm.until_next_phase("summon", lambda _: mj.x < 120)  # 走到二列停下来召唤吃1-1小喷
            await dm.until_next_phase("move", lambda _: iz_test.ground["1-1"] is None)  # 1-1死了收官

    @iz_test.on_game_end()
    def on_game_end(result):
        if not result:  # 过率是完爆, 死率疑似来源于后伴舞和mj一起走的时候步伐不同步导致mj被打(?
            print("mj is dead" if (z := iz_test.ground.zombie(0)) is None else (z.x, z.y))

    print(iz_test.start_test(jump_frame=jump_frame, speed_rate=2))



# -*- coding: utf_8 -*-
"""
验证安装用
"""
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="rpze command line utility")
    parser.add_argument("--path", help="specify a path for running the example")
    args = parser.parse_args()
    if p := args.path:
        print(f"your game path is {p}, "
              f"remember that only 1.0.0.1051_EN on pvz.tools is officially supported")
        try:
            from .basic.inject import InjectedGame
            from .examples.botanical_clock import botanical_clock
        except ImportError as ie:
            raise ImportError("maybe the package is not fully installed?") from ie

        from pathlib import Path
        dir_ = Path(__file__).parent / "bin"
        if not dir_.is_dir():
            raise IOError("/bin is not a directory, "
                          "please turn off the antivirus program and add exclusions.")
        file_names = set()
        for entry in dir_.iterdir():
            if entry.is_file():
                file_names.add(entry.name)
        if not {"rp_dll.dll", "rp_injector.exe"} <= file_names:
            raise IOError("miss binary dependencies! "
                          "please turn off the antivirus program and add exclusions.")
        try:
            game = InjectedGame(p)
        except (PermissionError, IOError, FileNotFoundError) as e:
            raise IOError("maybe the path is wrong?") from e
        with game:
            botanical_clock(game.controller, False)



# -*- coding: utf_8 -*-
"""
iztest 常见操作
"""
from .consts import plant_abbr_to_type, zombie_abbr_to_type
from ..basic.gridstr import parse_grid_str
from ..basic.inject import ConnectedContext, enter_level
from ..flow.utils import delay
from ..rp_extend import Controller
from ..structs.game_board import get_board, GameBoard
from ..structs.plant import Plant
from ..structs.zombie import Zombie


def enter_ize(ctler: Controller) -> GameBoard:
    """
    进入 ize 关卡并等待至清除选卡僵尸

    Args:
        ctler: 被注入的游戏对象
    Returns:
        进入关卡的 GameBoard 对象
    """
    with ConnectedContext(ctler) as ctler:
        enter_level(ctler, 70)
        board = get_board(ctler)
        board.remove_cutscene_zombie()
        while ctler.read_u32(0x6a9ec0, 0x7fc) == 2:  # game scene == playing
            ctler.skip_frames()
    return board


def place(place_str: str, board: GameBoard | None = None) -> Zombie | Plant | None:
    """
    用字符串放置植物

    Args:
        place_str: 放置植物/僵尸的字符串
        board: 要放置的board. 为None时使用get_board()
    Returns:
        放置的植物或者僵尸
    Raises:
        ValueError: 无法识别的植物或僵尸缩写
    Examples:
        >>> gb: GameBoard = ...
        >>> place("1 1-2", gb)
        放置一颗豌豆在1-2
        >>> place("cg 1-6")
        放置一个撑杆在1-6
    """
    if board is None:
        board = get_board()
    type_str, pos_str = place_str.split(" ")
    row, col = parse_grid_str(pos_str)
    if (type_ := zombie_abbr_to_type.get(type_str)) is not None:
        return board.iz_place_zombie(row, col, type_)
    if (type_ := plant_abbr_to_type.get(type_str)) is not None:
        return board.iz_new_plant(row, col, type_)
    raise ValueError(f"invalid type_str: {type_str}")


async def repeat(place_str: str,
                 time: int = 2,
                 interval: int = 20,
                 board: GameBoard | None = None) -> list[Zombie | Plant | None]:
    """
    生成一个连续放东西的flow

    Args:
        place_str: 放置植物/僵尸的字符串
        time: 放僵尸个数
        interval: 放僵尸间隔时间
        board: 要放置的 board. 默认使用 get_board()
    Returns:
        放置的植物或者僵尸列表
    Examples:
        >>> async def flow(_):
        ...    ...  # do something
        ...    await repeat("cg 1-6", time=3)
        为1-6三撑杆
    """
    ret = [place(place_str, board)]
    for _ in range(time - 1):
        await delay(interval)
        ret.append(place(place_str, board))
    return ret



# -*- coding: utf_8 -*-
"""
简化 iztest 编写的条件函数
"""
from typing import Literal

from ..flow.flow import FlowManager
from ..flow.utils import VariablePool, AwaitableCondFunc
from ..structs.plant import Plant, PlantStatus, PlantType


def until_precise_digger(magnet: Plant) -> AwaitableCondFunc[None]:
    """
    生成一个等到磁铁到达精确矿时间的函数

    Args:
        magnet: 要判断 cd 的磁铁
    """
    return AwaitableCondFunc(lambda _: magnet.status_cd <= 587)  # 587 - 590 by 寒风


def until_plant_die(plant: Plant) -> AwaitableCondFunc[None]:
    """
    生成一个等到植物死亡的函数

    Args:
        plant: 要判断的植物
    """
    return AwaitableCondFunc(lambda _: plant.is_dead)


def until_plant_last_shoot(plant: Plant, wait_until_mbd: bool = False) -> AwaitableCondFunc[int]:
    """
    生成一个 等到植物 "本段最后一次连续攻击结束后返回" 的函数.

    await 调用后返回"开打帧距离上一次攻击的距离"

    对裂荚处理逻辑与对其他植物有区别: 只关心第一发. 裂荚右与单发结果相同, 裂荚左与双发结果不同.

    Args:
        plant: 要判断的植物
        wait_until_mbd: 是否等到 上次开打经过最大攻击间隔后 再返回
    Examples:
        >>> async def flow(_):
        ...     plant = iz_test.ground["1-2"]  # noqa
        ...     t = await until_plant_last_shoot(plant)
        ...     assert plant.max_boot_delay - 14 <= t <= plant.max_boot_delay  # t 即为攻击间隔时长
    """

    shoot_next_gcd = 1 if plant.type_ is not PlantType.split_pea else 26  # 修正裂荚攻击时机
    mbd = plant.max_boot_delay

    def _await_func(fm: FlowManager, v=VariablePool(
            try_to_shoot_time=None,
            last_shooting_time=None,
            until_mbd_ret=None)):
        if v.until_mbd_ret is not None:  # until mbd flag开了就走: 等到最大攻击间隔后再返回
            if fm.time >= v.last_shooting_time + mbd:
                return True, v.until_mbd_ret
            return False
        if plant.generate_cd == shoot_next_gcd:  # 下一帧开打
            v.try_to_shoot_time = fm.time + 1
        if v.try_to_shoot_time == fm.time:
            if plant.launch_cd > 15:  # 判断大于15则处于攻击状态, 目的是兼容忧郁菇
                # 一般植物处于非攻击状态launch_cd为0, 忧郁菇处于非攻击状态的launch_cd最大值为14
                v.last_shooting_time = fm.time
                return False
            else:  # 不处于攻击状态
                if v.last_shooting_time is not None:
                    if not wait_until_mbd or fm.time == v.last_shooting_time + mbd:
                        return True, fm.time - v.last_shooting_time
                    # 如果等最大攻击间隔再返回 flag改not None开始走until逻辑
                    v.until_mbd_ret = fm.time - v.last_shooting_time
                    return False
                v.last_shooting_time = None
                return False  # 上一轮是攻击的 且 这一轮不攻击 返回True
        return False

    return AwaitableCondFunc(_await_func)


def until_plant_n_shoot(plant: Plant, n: int = 1, non_stop: bool = True) -> AwaitableCondFunc[None]:
    """
    生成一个 等到植物n次攻击 的函数

    Args:
        plant: 要判断的植物
        n: 攻击次数
        non_stop: 是否为不间断攻击
    """

    shoot_next_gcd = 1 if plant.type_ is not PlantType.split_pea else 26  # 修正裂荚攻击时机
    
    def _await_func(fm: FlowManager,
                    v=VariablePool(try_to_shoot_time=None, shots=0)):
        if plant.generate_cd == shoot_next_gcd:  # 下一帧开打
            v.try_to_shoot_time = fm.time + 1
        if v.try_to_shoot_time == fm.time:
            if plant.launch_cd > 15:  # 在攻击时
                v.shots += 1
            else:  # 不再攻击时
                if non_stop:  # 设置了不停止标志，则计数清零
                    v.shots = 0
        if v.shots == n:
            return True
        return False
    
    return AwaitableCondFunc(_await_func)


CountButterModeLiteral = Literal[0, 1, 2, "total", "nonstop", "continuous"]
"""数黄油函数 until_n_butter 的计数方法

    - 0 或 "total" 表示允许攻击中断
    - 1 或 "nonstop" 表示攻击不中断
    - 2 或 "continuous" 表示攻击不中断, 而且黄油必须连续投出
"""


def until_n_butter(plant: Plant, n: int = 1, mode: CountButterModeLiteral = 1) -> AwaitableCondFunc[None]:
    """
    生成一个 等到玉米攻击n发黄油 的函数

    Args:
        plant: 要判断的植物
        n: 攻击黄油次数
        mode: 字面量, 表示计数方法
    """
    match mode:
        case "total" | 0:
            mode_index = 0
        case "nonstop" | 1:
            mode_index = 1
        case "continuous" | 2:
            mode_index = 2
        case _:
            raise ValueError(f"invalid count mode: {mode}")

    def _await_func(fm: FlowManager, v=VariablePool(projs=0, try_to_shoot_time=None)):
        if plant.generate_cd == 1:  # 下一帧开打
            v.try_to_shoot_time = fm.time + 1
        if v.try_to_shoot_time == fm.time:
            if plant.status is PlantStatus.kernelpult_launch_butter:  # 出黄油
                v.projs += 1
            elif plant.launch_cd == 0:  # 攻击停止
                if mode_index != 0:
                    v.projs = 0
            else:  # 出玉米粒
                if mode_index == 2:
                    v.projs = 0
        if v.projs == n:
            return True 
        return False
    
    return AwaitableCondFunc(_await_func)



# -*- coding: utf_8 -*-
"""
ize 对植物常见的操控方式
"""
import random
from collections.abc import Iterable
from typing import overload

from ..structs.game_board import get_board
from ..structs.plant import Plant, PlantType


def randomize_generate_cd(plant: Plant) -> Plant:
    """
    令植物的 generate_cd 按照"放置充分长时间"后的结果随机化

    **仅对can_attack == True植物有效**; 但特判地刺, 地刺王无效.

    具体来说, 其 generate_cd 概率分布图像为一个梯形:
    上底为 max_boot_delay - 14, 下底为 max_boot_delay.

    Returns:
        返回传入的植物
    """
    if (not plant.can_attack) or plant.type_ in {PlantType.spikeweed, PlantType.spikerock}:
        return plant
    # 拆成[1, max_ - 14)和[max_ - 14, max_ + 1)两个区间
    # 不可以取0, 可以取max_, max_ - 14和前面概率相等为h
    # h * (max_ - 15) + (h + 0) * 16 / 2 = 1解这个方程, h为梯形的高
    h = 1 / ((max_ := plant.max_boot_delay) - 7)
    distribution = [h] * (max_ - 15) + [h / 15 * i for i in range(15, 0, -1)]
    plant.generate_cd = random.choices(population=range(1, max_ + 1), weights=distribution)[0]
    return plant


@overload
def set_puff_x_offset(puff: Plant, offset: int) -> None:
    """
    为小喷设置 x 偏移

    Args:
        puff: 目标小喷
        offset: 小喷 x 偏移
    Raises:
        ValueError: offset不在范围内
    Examples:
        >>> p = ...
        >>> set_puff_x_offset(p, 3)
        为小喷设置x偏移为+3
    """


@overload
def set_puff_x_offset(puff: Plant, offset: Iterable[int]) -> None:
    """
    为小喷设置 x 偏移

    Args:
        puff: 目标小喷
        offset: 小喷 x 偏移范围
    Raises:
        ValueError: **最终随机结果的**偏移不在范围内.
    Examples:
        >>> p = ...
        >>> set_puff_x_offset(p, range(-5, 4))
        为小喷设置在整个范围内的随机x偏移
    """


def set_puff_x_offset(puff: Plant, offset):
    center_x = get_board(puff.controller).grid_to_pixel_x(puff.col, puff.row)
    offset = offset if isinstance(offset, int) else random.choice(tuple(offset))
    if not (-5 <= offset <= 4):
        raise ValueError(f"offset {offset} out of valid range of puff")
    puff.x = center_x + offset



# -*- coding: utf_8 -*-
"""
流程控制相关的函数和类
"""
from collections.abc import Callable, Coroutine
from enum import Enum, auto
from itertools import count
from typing import TypeAlias, Self, Final, Any


class TickRunnerResult(Enum):
    """ TickRunner 的返回值"""
    DONE = auto()
    """本 TickRunner 以后再也不执行时返回"""
    BREAK_DONE = auto()  # 不用异常打断. StopIteration形式的返回值和type hint系统匹配程度太差.
    """需要打断本次运行并且以后不再运行本 TickRunner 时返回"""
    BREAK_ONCE = auto()
    """需要打断本次运行但以后还会运行本 TickRunner 时返回"""


CondFunc: TypeAlias = Callable[["FlowManager"], bool]
"""判断条件的函数"""
AwaitFunc: TypeAlias = Callable[["FlowManager"], bool | tuple[bool, *tuple[Any, ...]]]
"""await 语句返回的函数, rets should be False / True / (True, *rets)"""
FlowCoroutine: TypeAlias = Coroutine[AwaitFunc, Any, TickRunnerResult | None]
"""Flow 返回的协程对象"""
Flow: TypeAlias = Callable[["FlowManager"], FlowCoroutine]
"""await AwaitFunc 函数的 async def 函数"""
TickRunner: TypeAlias = Callable[["FlowManager"], TickRunnerResult | None]
"""帧运行函数, 无返回值表示继续执行, 返回 TickRunnerResult 表示特殊行为"""


def split_await_func_ret(rets: bool | tuple[bool, *tuple[Any, ...]]) -> tuple[bool, Any]:
    """
    把 AwaitFunc 返回值拆成 bool / others 两份.

    Args:
        rets: AwaitFunc 返回值
    Returns:
        AwaitFunc -> bool 时, 返回 bool, None
        AwaitFunc -> tuple[bool, T] 时, 返回 bool, T;
        AwaitFunc -> tuple[bool, *Args] 时, 返回 bool, tuple[*Args]
    """
    match rets:  # rets should be False / True / (True, *rets)
        case (_, _):
            return rets
        case (b, *args):
            return b, tuple(args)
        case b:
            return b, None


class FlowManager:
    """
    运行 Flow 和 TickRunner 函数的对象

    Attributes:
        tick_runners: 所有 TickRunner 组成的函数, 运行时按顺序执行
        destructors: 所有析构函数组成的函数, 在 end() 时按顺序执行
        time: 每执行一次 do() 自增 1
    """

    def __init__(self,
                 tick_runners: list[tuple[int, TickRunner]],
                 destructors: list[tuple[int, Callable[[Self], None]]],
                 flows: list[Flow],
                 flow_priority: int,
                 flow_destructor_priority: int):
        """
        Args:
            tick_runners: tick_runner 列表, 以(priority, tick_runner)形式提供以便排序
            destructors: 析构列表, 在调用 self.end() 时执行, 以(priority, func)形式提供以便排序
            flows: flow 列表
            flow_priority: flows 执行优先级
            flow_destructor_priority: flows 析构优先级
        """
        self.time = 0
        self._is_destructed = False
        self._flow_coro_list: list[list] = [[lambda _: True, i(self)] for i in flows]

        def _flow_tick_runner(self_: "FlowManager") -> TickRunnerResult | None:
            if not (fcl := self_._flow_coro_list):
                return TickRunnerResult.DONE
            pop_list = []
            for idx, (await_func, flow) in enumerate(fcl):
                try:
                    rets = await_func(self_)
                except BaseException as e:
                    flow.throw(e)
                    continue  # 如果能处理异常则继续执行
                b, ret = split_await_func_ret(rets)
                if b:
                    try:
                        fcl[idx][0] = flow.send(ret)
                    except StopIteration as se:  # StopIteration.value为返回值
                        pop_list.append(idx)
                        if se.value is None:
                            continue
                        for i in pop_list[::-1]:
                            fcl.pop(i)
                        return se.value
            for i in pop_list[::-1]:
                fcl.pop(i)

        _counter = count()
        tick_runner_list = [(-priority, next(_counter), it) for priority, it in tick_runners]
        tick_runner_list.append((-flow_priority, next(_counter), _flow_tick_runner))
        tick_runner_list.sort()
        # -priority让priority越大优先级别越高
        self.tick_runners: list[TickRunner] = [i[2] for i in tick_runner_list]

        def _flow_destructor(self_: "FlowManager"):
            for flow in self_._flow_coro_list:
                flow[1].close()

        destructor_list = [(-priority, next(_counter), it) for priority, it in destructors]
        destructor_list.append((-flow_destructor_priority, next(_counter), _flow_destructor))
        destructor_list.sort()
        self.destructors: list[Callable[["FlowManager"], None]] = [i[2] for i in destructor_list]

    def add(self) -> Callable[[TickRunner], TickRunner]:
        """
        运行时添加 TickRunner 的装饰器

        运行时添加的 TickRunner 会被放在最后执行. 即, 不支持加优先级, 但确保本帧执行.

        Examples:
            >>> flow_manager: FlowManager = ...
            >>> @flow_manager.add()
            ... def tr(fm: "FlowManager") -> TickRunnerResult:
            ...     ...
            为装饰器形式使用

            >>> flow_manager: "FlowManager" = ...
            >>> def tr(fm: "FlowManager") -> TickRunnerResult:
            ...     ...
            >>> flow_manager.add()(tr)
            为函数形式使用
        """

        def _decorator(tr: TickRunner):
            self.tick_runners.append(tr)
            return tr

        return _decorator

    def add_destructor(self) \
            -> Callable[[Callable[["FlowManager"], None]], Callable[["FlowManager"], None]]:
        """
        添加 destructor 的方法, 与 FlowManager.add 使用方法相同
        """

        def _decorator(d: Callable[["FlowManager"], None]) -> Callable[["FlowManager"], None]:
            self.destructors.append(d)
            return d

        return _decorator

    def connect(self, cond: CondFunc, only_once: bool = False) \
            -> Callable[[TickRunner], TickRunner]:
        """
        运行时把 tick_runner 绑定到 cond 上的方法, 与 add 使用方法相同
        
        即 在 cond(self) 返回 True 时执行 func(self).

        Args:
            cond: 执行 func 的条件函数. 返回 None 时按照 only_once 判断; 返回 TickRunnerResult 时直接返回.
            only_once: 为 True 时, 只要有一次满足 cond 则返回
        """

        def _decorator(tr: TickRunner) -> TickRunner:
            def _decorated_tick_runner(fm: "FlowManager"):
                if cond(fm):
                    ret = tr(fm)
                    if ret is None:
                        return TickRunnerResult.DONE if only_once else None
                    return ret

            self.add()(_decorated_tick_runner)
            return tr

        return _decorator

    def run(self) -> TickRunnerResult | None:
        """
        运行一次内部所有函数

        Returns:
            所有 tick_runner 都执行完毕或对象已经析构时返回 DONE, 内部有人打断时返回 BREAK_ONCE, 否则返回空.
        """
        if not (trs := self.tick_runners) or self._is_destructed:
            return TickRunnerResult.DONE
        pop_list = []

        def _end(_type):
            self.time += 1
            for it in pop_list[::-1]:
                trs.pop(it)
            return _type

        for idx, func in enumerate(trs):
            ret: TickRunnerResult | None = func(self)
            match ret:
                case None:
                    continue
                case TickRunnerResult.DONE:
                    pop_list.append(idx)  # 早该换成链表了
                case TickRunnerResult.BREAK_DONE:
                    pop_list.append(idx)
                    return _end(TickRunnerResult.BREAK_ONCE)
                case TickRunnerResult.BREAK_ONCE:
                    return _end(TickRunnerResult.BREAK_ONCE)
        return _end(None)

    def end(self) -> None:
        """
        结束运行时执行析构函数
        """
        self._is_destructed = True
        for d in self.destructors:
            d(self)


DEFAULT_PRIORITY: Final[int] = 0
"""各类参数的默认优先级"""


class FlowFactory:
    """
    用于生成 FlowManager 的工厂对象

    Attributes:
        flow_list: 所有 flow 组成的列表
        tick_runner_list: 所有 tick_runner 组成的列表
        destructor_list: 在 FlowManager 结束运行后会执行的所有函数
    """

    def __init__(self):
        self.flow_list: list[Flow] = []
        self.tick_runner_list: list[tuple[int, TickRunner]] = []
        self.destructor_list: list[tuple[int, Callable[[FlowManager], None]]] = []

    def add_flow(self) -> Callable[[Flow], Flow]:
        """
        添加 flow 的方法, 与 FlowManager.add 使用方法相同
        """

        def _decorator(f: Flow) -> Flow:
            self.flow_list.append(f)
            return f

        return _decorator

    def add_tick_runner(self, priority: int = DEFAULT_PRIORITY) \
            -> Callable[[TickRunner], TickRunner]:
        """
        添加 tick_runner 的方法, 与 FlowManager.add 使用方法相同

        Args:
            priority: 权重 越大越优先执行
        """

        def _decorator(tr: TickRunner):
            self.tick_runner_list.append((priority, tr))
            return tr

        return _decorator

    def add_destructor(self, priority: int = DEFAULT_PRIORITY) \
            -> Callable[[Callable[[FlowManager], None]], Callable[[FlowManager], None]]:
        """
        添加 destructor 的方法, 与 FlowManager.add 使用方法相同

        Args:
            priority: 权重 越大越优先执行
        """

        def _decorator(d: Callable[[FlowManager], None]) -> Callable[[FlowManager], None]:
            self.destructor_list.append((priority, d))
            return d

        return _decorator

    def connect(self, cond: CondFunc, priority: int = DEFAULT_PRIORITY, only_once: bool = False) \
            -> Callable[[TickRunner], TickRunner]:
        """
        把 tick_runner 绑定到 cond 上的方法, 与 FlowManager.add 使用方法相同

        Args:
            cond: 执行 func 的条件函数. 返回 None 时按照 only_once 判断; 返回 TickRunnerResult 时直接返回.
            priority: 权重 越大越优先执行
            only_once: 为 True 时 仅当第一次满足 cond 时执行
        """

        def _decorator(tr: TickRunner) -> TickRunner:
            def _decorated_tick_runner(fm: FlowManager):
                if cond(fm):
                    ret = tr(fm)
                    if ret is None:
                        return TickRunnerResult.DONE if only_once else None
                    return ret

            self.add_tick_runner(priority)(_decorated_tick_runner)
            return tr

        return _decorator

    def build_manager(self, flow_priority: int = DEFAULT_PRIORITY,
                      flow_destructor_priority: int = DEFAULT_PRIORITY) -> FlowManager:
        """
        生成 FlowManager 的方法

        Args:
            flow_priority: flow 在 tick runner 中的权重, 越大越优先执行
            flow_destructor_priority: flow 析构函数在析构函数中的权重, 越大越优先执行
        Returns:
            生成的 FlowManager 对象
        """
        return FlowManager(self.tick_runner_list, self.destructor_list, self.flow_list,
                           flow_priority, flow_destructor_priority)



# -*- coding: utf_8 -*-
"""
ize 测试功能示例; everything in this package have no compability
"""
from .botanical_clock import botanical_clock  # noqa
from .dancing_example import dancing_example  # noqa
from .end_callback_example import end_test  # noqa
from .pole_jumping_test import pole_jumping_test  # noqa
from .iztools_example import default_test  # noqa



# -*- coding: utf_8 -*-
# extend module of game controller
from enum import Enum
from typing import Self


class HookPosition(Enum):
    MAIN_LOOP = 0
    # the main hook of the project
    # do not open / close it manually, use Controller.start()/.end() instead
    ZOMBIE_PICK_RANDOM_SPEED = 1  # useless
    CHALLENGE_I_ZOMBIE_SCORE_BRAIN = 2
    # open to disable 0x42B8B0 (disable := do nothing and return)
    CHALLENGE_I_ZOMBIE_PLACE_PLANTS = 3
    # open to disable 0x42A6C0


class SyncMethod(Enum):
    SPIN = 1  # better performance for testing
    MUTEX = 2  # better performance for normal case like modifier


class RpBaseException(Exception): ...


class ControllerError(RpBaseException): ...


class Controller:
    # game controller

    # in the description of this class:
    #     connected := using start(), which means hook_connected(HookPosition.MAIN_LOOP) is True
    #     prepared := hook connected and before() used properly

    def __init__(self, pid: int, /) -> None: ...

    def __eq__(self, other: Self, /) -> bool: ...

    def __ne__(self, other: Self, /) -> bool: ...

    def __repr__(self) -> str: ...

    @property
    def result_mem(self) -> memoryview: ...  # a free piece of shared memory

    @property
    def result_address(self) -> int: ...  # the address of result_mem in game

    # value of result_mem, same as *(T*)result_address
    result_bool: bool
    result_f32: float
    result_f64: float
    result_i8: int
    result_i16: int
    result_i32: int
    result_i64: int
    result_u8: int
    result_u16: int
    result_u32: int
    result_u64: int

    sync_method: SyncMethod  # default: SyncMethod.MUTEX, cannot change when connected
    jumping_sync_method: SyncMethod  # default: SyncMethod.SPIN , cannot change when jumping frame

    @property
    def asm_address(self) -> int: ...  # start addr of where run_code would be executed

    @property
    def pid(self) -> int: ...  # game process id

    def skip_frames(self, num: int = 1) -> None: ...  # assert prepared; skip {num} frames and get prepared

    def before(self) -> None: ...  # required before every frame after start() control

    def next_frame(self) -> None: ...  # let the game continue to the next frame

    def start(self) -> None: ...  # start control and get prepared; do nothing when connected

    def end(self) -> None: ...
    # assert prepared; end control (end jumping frame if necessary); do nothing when not connected

    def start_jump_frame(self) -> bool: ...
    # assert prepared and has board
    # return False if already jumping

    def is_jumping_frame(self) -> bool: ...

    def end_jump_frame(self) -> bool: ...  # assert prepared; return False if not jumping, skip a frame

    def get_p_board(self) -> tuple[bool, int]: ...  # return (is_p_board_new, p_board)

    def run_code(self, asm_bytes: bytes, /) -> bool: ...  # assert prepared; return False if failed

    def open_hook(self, hook: HookPosition, /) -> None: ...

    def close_hook(self, hook: HookPosition, /) -> None: ...

    def hook_connected(self, hook: HookPosition = HookPosition.MAIN_LOOP) -> bool: ...

    def global_connected(self) -> bool: ...  # return False if game is closed

    # fall back to use ReadProcessMemory & WriteProcessMemory when force_remote or not_prepared
    # *args for offsets
    # return None(read_... funcs) or False(write_... funcs) when reading nullptr
    # the game will crash when reading wild pointers
    def read_bool(self, *args: int, force_remote: bool = False) -> bool | None: ...

    def read_f32(self, *args: int, force_remote: bool = False) -> float | None: ...

    def read_f64(self, *args: int, force_remote: bool = False) -> float | None: ...

    def read_i8(self, *args: int, force_remote: bool = False) -> int | None: ...

    def read_i16(self, *args: int, force_remote: bool = False) -> int | None: ...

    def read_i32(self, *args: int, force_remote: bool = False) -> int | None: ...

    def read_i64(self, *args: int, force_remote: bool = False) -> int | None: ...

    def read_u8(self, *args: int, force_remote: bool = False) -> int | None: ...

    def read_u16(self, *args: int, force_remote: bool = False) -> int | None: ...

    def read_u32(self, *args: int, force_remote: bool = False) -> int | None: ...

    def read_u64(self, *args: int, force_remote: bool = False) -> int | None: ...

    def read_bytes(self, size: int, *args: int, force_remote: bool = False) -> bytes | None: ...


    def write_bool(self, value: bool, *args: int, force_remote: bool = False) -> bool: ...

    def write_f32(self, value: float, *args: int, force_remote: bool = False) -> bool: ...

    def write_f64(self, value: float, *args: int, force_remote: bool = False) -> bool: ...

    def write_i8(self, value: int, *args: int, force_remote: bool = False) -> bool: ...

    def write_i16(self, value: int, *args: int, force_remote: bool = False) -> bool: ...

    def write_i32(self, value: int, *args: int, force_remote: bool = False) -> bool: ...

    def write_i64(self, value: int, *args: int, force_remote: bool = False) -> bool: ...

    def write_u8(self, value: int, *args: int, force_remote: bool = False) -> bool: ...

    def write_u16(self, value: int, *args: int, force_remote: bool = False) -> bool: ...

    def write_u32(self, value: int, *args: int, force_remote: bool = False) -> bool: ...

    def write_u64(self, value: int, *args: int, force_remote: bool = False) -> bool: ...

    def write_bytes(self, value: bytes, *args: int, force_remote: bool = False) -> bool: ...



# -*- coding: utf_8 -*-
"""
ize测试功能
"""
from ..basic.inject import InjectedGame  # noqa
from ..flow.utils import until, delay, VariablePool  # noqa
from ..rp_extend import Controller  # noqa
from ..structs.zombie import ZombieStatus, ZombieType, Zombie  # noqa
from ..structs.plant import PlantStatus, PlantType, Plant  # noqa
from .iztest import IzTest  # noqa
from .operations import *  # noqa
from .cond_funcs import *  # noqa



# -*- coding: utf_8 -*-
"""
简化flow编写的工具函数
"""
import warnings
from collections.abc import Awaitable, Generator
from typing import overload, Self, Any, TypeVar

from .flow import CondFunc, FlowManager, AwaitFunc, split_await_func_ret


class VariablePool:  # thanks Reisen
    """
    用于表示 AwaitFunc 中用于"伴随状态"的默认参数的变量池.
    具体属性由构造函数而定
    """

    def reset(self) -> Self:
        """
        重置变量池. 重置后变量池的属性值为初始值.

        Returns:
            self
        """
        self._args_list = list(self.origin_list)
        for k, v in self.origin_dict.items():
            setattr(self, k, v)
        return self

    def __init__(self, *args, **kwargs):
        """
        Args:
            *args: 匿名变量. 使用下标运算访问.
            **kwargs: 属性. 属性名=初始值
        """
        self._args_list = []
        self.origin_list = args
        self.origin_dict = kwargs
        self.reset()

    def __getitem__(self, item):
        return self._args_list[item]

    def __setitem__(self, key, value):
        self._args_list[key] = value

    def get_all_attrs(self) -> tuple[tuple, dict] | None:
        """
        获取所有属性的值

        Returns:
            所有匿名属性(in a tuple), 所有具名属性(in a dict)
        """
        return tuple(self._args_list), {k: getattr(self, k) for k in self.origin_dict}

    def __str__(self):
        t, d = self.get_all_attrs()
        return f"<{t}, {d}>"


_T_co = TypeVar("_T_co", covariant=True)


def _await_generator(t) -> Generator[Any, Any, _T_co]:
    ret = yield t
    return ret


class AwaitableCondFunc(Awaitable[_T_co], CondFunc):
    """
    包装 Func 对象用于在 flow 中 await 调用.

    泛型变量用于表示: reveal_type(await AwaitableCondFunc[T]) is T

    Attributes:
        func: 内层 CondFunc 函数
    """
    __slots__ = ("func",)

    def __init__(self, func: AwaitFunc):
        self.func: AwaitFunc = func

    def __call__(self, fm: FlowManager) -> bool:
        """
        调用内层 func. 确保 AwaitableCondFunc 自己也为 CondFunc 函数.

        忽略可能作为 await 返回值的 _T_co 而只返回第一位 bool.
        """
        return split_await_func_ret(self.func(fm))[0]

    def __await__(self) -> Generator[AwaitFunc, Any, _T_co]:
        """
        让 AwaitableCondFunc 对象可以 await.

        Returns:
            生成器对象. 唯一一个生成结果为 self.func.
        """
        return _await_generator(self.func)

    def __and__(self, other: CondFunc) -> "AwaitableCondFunc[None]":
        """
        重载&运算符, 使得对象可以用&运算符, 像逻辑和运算一样连接

        Args:
            other: 另一个 CondFunc 对象

        Returns:
            一个新的 AwaitableCondFunc 对象. 该对象的func为self() and other()
        """
        return AwaitableCondFunc(lambda fm: self(fm) and other(fm))

    def __rand__(self, other: CondFunc) -> "AwaitableCondFunc[None]":
        """
        重载&运算符, 使得对象可以用&运算符, 像逻辑和运算一样连接

        Args:
            other: 另一个 CondFunc 对象

        Returns:
            一个新的 AwaitableCondFunc 对象. 该对象的 func 为 other() and self()
        """
        return AwaitableCondFunc(lambda fm: other(fm) and self(fm))

    def __or__(self, other: CondFunc) -> "AwaitableCondFunc[None]":
        """
        重载|运算符, 使得对象可以用|运算符, 像逻辑或运算一样连接

        Args:
            other: 另一个CondFunc对象
        Returns:
            一个新的 AwaitableCondFunc 对象. 该对象的 func 为 self() or other()
        """
        return AwaitableCondFunc(lambda fm: self(fm) or other(fm))

    def __ror__(self, other: CondFunc) -> "AwaitableCondFunc[None]":
        """
        重载|运算符, 使得对象可以用|运算符, 像逻辑或运算一样连接

        Args:
            other: 另一个 CondFunc 对象
        Returns:
            一个新的 AwaitableCondFunc 对象. 该对象的 func 为 other.func or self()
        """
        return AwaitableCondFunc(lambda fm: other(fm) or self(fm))

    def __invert__(self) -> "AwaitableCondFunc[None]":
        """
        重载~运算符, 使得对象可以用~运算符, 像逻辑非运算一样.

        Returns:
            一个新的 AwaitableCondFunc 对象. 该对象的 func 为 not self()
        """
        return AwaitableCondFunc(lambda fm: not self(fm))

    def after(self, delay_time: int) -> Self:
        """
        生成一个 在满足原条件后过 delay_time 帧返回原返回值的对象.
        Args:
            delay_time: 延迟时间
        Returns:
            一个新的 AwaitableCondFunc 对象.
        """

        def _await_func(fm: FlowManager, p=VariablePool(end_time=None, ret=None)):
            if p.end_time is None:  # end_time未赋值即self.func未返回True
                b, ret = split_await_func_ret(self.func(fm))
                if b:
                    p.end_time = fm.time + delay_time
                    p.ret = ret
            if p.end_time is not None and p.end_time <= fm.time:
                return True, p.ret  # end_time赋值则按时间返回
            return False

        return AwaitableCondFunc(_await_func)


@overload
def until(time: int, /) -> AwaitableCondFunc[None]:
    """
    生成一个 判断时间是否到达 的函数

    Args:
        time: 当前时间大于等于time时返回True
    Examples:
        >>> async def flow(_):
        ...     await until(100)
        ...     ...  # do something
        为一个 在time >= 100时执行do something的flow
    """


@overload
def until(cond_func: CondFunc, /) -> AwaitableCondFunc[None]:
    """
    把cond_func函数包装为AwaitableCondFunc对象.

    Args:
        cond_func: 判断条件的函数
    Returns:
        一个包装的AwaitableCondFunc函数.
    """


def until(arg) -> AwaitableCondFunc[None]:
    if callable(arg):
        return AwaitableCondFunc(arg)
    if isinstance(arg, bool):
        warnings.warn(
            "until(bool) is usually not what you want, use until(lambda _: bool) instead.",
            SyntaxWarning, stacklevel=2)
    return AwaitableCondFunc(lambda fm: fm.time >= arg)


def delay(time: int) -> AwaitableCondFunc[None]:
    """
    生成一个 延迟time帧后返回True的函数

    Args:
        time: 延迟用时间
    Raises:
        ValueError: time <= 0时候抛出.
    Examples:
        >>> async def flow(_):
        ...     ...  # do something
        ...     await delay(50)
        ...     ...  # do other thing
        为相隔50cs连续执行
    """
    if time <= 0:
        raise ValueError(f"time must be positive, not {time}")

    def _cond_func(fm, v=VariablePool(start_time=None)) -> bool:
        if v.start_time is None:
            v.start_time = fm.time
        if v.start_time + time - 1 <= fm.time:  # 所有CondFunc函数下一cs开始执行.
            return True
        return False

    return AwaitableCondFunc(_cond_func)



# -*- coding: utf_8 -*-
"""
描述 pvz 中数据结构的基类和基本函数.
"""
import abc
from collections.abc import Sequence, Iterator, Callable
from enum import IntEnum
from typing import ClassVar, Self, TypeVar, overload, SupportsIndex, Generic, Any, TypeAlias, Final, SupportsInt

from ..basic import asm
from ..basic.exception import PvzStatusError
from ..rp_extend import Controller


class ObjBase(abc.ABC):
    """
    pvz 中的一个对象

    Attributes:
        base_ptr: 对应 pvz 中对象的指针
        controller: 对应 pvz 的 Controller 对象
    """
    __slots__ = ("base_ptr", "controller")

    def __init__(self, base_ptr: int, ctler: Controller):
        """
        一个 ObjBase 对象由一个指向 pvz 中的对象的指针, 和对应游戏的 Controller 构造
        
        Args:
            base_ptr: 游戏中对象的基址
            ctler: 游戏对应的 Controller 对象
        Raises:
            ValueError: base_ptr 为空时抛出
        """
        if base_ptr == 0:
            raise ValueError(f"base_ptr of an {type(self).__name__} object cannot be 0")
        super().__init__()
        self.base_ptr = base_ptr
        self.controller = ctler

    OBJ_SIZE: ClassVar[int] = NotImplemented
    """对应 pvz 类在 pvz 中的大小, 必须在所有非抽象子类中赋值"""

    def __eq__(self, other: Self) -> bool:
        """
        判断二个 ObjBase 对象是否指向同一游戏的同一位置

        功能更接近于 Python 中的 is.

        Args:
            other : 另一个 ObjBase 对象
        """
        return self.base_ptr == other.base_ptr and (self.controller == other.controller)

    def __ne__(self, other: Self) -> bool:
        return not (self.base_ptr == other.base_ptr and (self.controller == other.controller))

    def __str__(self) -> str:
        return (f"<{type(self).__name__} object at [0x{self.base_ptr:x}] "
                f"of process id {self.controller.pid}>")

    def __repr__(self) -> str:
        return (f"{type(self).__name__}(base_ptr=0x{self.base_ptr:x}, "
                f"ctler=Controller({self.controller.pid}))")


_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_T_con = TypeVar("_T_con", contravariant=True)


class OffsetProperty(property, Generic[_T_co, _T_con]):
    """
    ObjBase对象在 pvz 内的属性

    两个泛型参数分别表示: 通过对象调用的 __get__ 返回值, __set__ 入参.

    Attributes:
        offset: 属性在游戏中的偏移
    """
    def __init__(self,
                 fget: Callable[[ObjBase], _T_co] | None,
                 fset: Callable[[ObjBase, _T_con], None] | None,
                 fdel: Callable[[ObjBase], None] | None,
                 doc: str | None,
                 offset: int):
        super().__init__(fget, fset, fdel, doc)
        self.__doc__ = doc
        self.__objclass__ = ObjBase
        self.offset: Final[int] = offset

    @overload
    def __get__(self, obj: None, owner: type | None = ..., /) -> Self: ...

    @overload
    def __get__(self, obj: Any, owner: type | None = ..., /) -> _T_co: ...

    def __get__(self, *args):
        return super().__get__(*args)

    def __set__(self, obj: ObjBase, value: _T_con) -> None:
        return super().__set__(obj, value)

    def __delete__(self, obj: ObjBase) -> None:
        return super().__delete__(obj)


_OffsetProp: TypeAlias = OffsetProperty[_T, _T]


# property factories 用于生成 ObjBase 对象在 pvz 内的 attribute
def property_bool(offset: int, doc: str) -> _OffsetProp[bool]:
    def _get(self: ObjBase) -> bool:
        return self.controller.read_bool(self.base_ptr + offset)

    def _set(self: ObjBase, value: bool):
        self.controller.write_bool(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "bool: " + doc, offset)


def property_i8(offset: int, doc: str) -> _OffsetProp[int]:
    def _get(self: ObjBase) -> int:
        return self.controller.read_i8(self.base_ptr + offset)

    def _set(self: ObjBase, value: int):
        self.controller.write_i8(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_i16(offset: int, doc: str) -> _OffsetProp[int]:
    def _get(self: ObjBase) -> int:
        return self.controller.read_i16(self.base_ptr + offset)

    def _set(self: ObjBase, value: int):
        self.controller.write_i16(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_i32(offset: int, doc: str) -> _OffsetProp[int]:
    def _get(self: ObjBase) -> int:
        return self.controller.read_i32(self.base_ptr + offset)

    def _set(self: ObjBase, value: int):
        self.controller.write_i32(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_i64(offset: int, doc: str) -> _OffsetProp[int]:
    def _get(self: ObjBase) -> int:
        return self.controller.read_i64(self.base_ptr + offset)

    def _set(self: ObjBase, value: int):
        self.controller.write_i64(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_u8(offset: int, doc: str) -> _OffsetProp[int]:
    def _get(self: ObjBase) -> int:
        return self.controller.read_u8(self.base_ptr + offset)

    def _set(self: ObjBase, value: int):
        self.controller.write_u8(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_u16(offset: int, doc: str) -> _OffsetProp[int]:
    def _get(self: ObjBase) -> int:
        return self.controller.read_u16(self.base_ptr + offset)

    def _set(self: ObjBase, value: int):
        self.controller.write_u16(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_u32(offset: int, doc: str) -> _OffsetProp[int]:
    def _get(self: ObjBase) -> int:
        return self.controller.read_u32(self.base_ptr + offset)

    def _set(self: ObjBase, value: int):
        self.controller.write_u32(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_u64(offset: int, doc: str) -> _OffsetProp[int]:
    def _get(self: ObjBase) -> int:
        return self.controller.read_u64(self.base_ptr + offset)

    def _set(self: ObjBase, value: int):
        self.controller.write_u64(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_f32(offset: int, doc: str) -> _OffsetProp[float]:
    def _get(self: ObjBase) -> float:
        return self.controller.read_f32(self.base_ptr + offset)

    def _set(self: ObjBase, value: float):
        self.controller.write_f32(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "float: " + doc, offset)


def property_f64(offset: int, doc: str) -> _OffsetProp[float]:
    def _get(self: ObjBase) -> float:
        return self.controller.read_f64(self.base_ptr + offset)

    def _set(self: ObjBase, value: float):
        self.controller.write_f64(value, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, "float: " + doc, offset)


_T_int_enum = TypeVar("_T_int_enum", bound=IntEnum)


def property_int_enum(offset: int, cls: type[_T_int_enum], doc: str) \
        -> OffsetProperty[_T_int_enum, SupportsInt]:
    def _get(self: ObjBase) -> _T_int_enum:
        return cls(self.controller.read_i32(self.base_ptr + offset))

    def _set(self: ObjBase, value: SupportsInt) -> None:
        self.controller.write_i32(int(value), self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, f"{cls.__name__}: {doc}", offset)


_T_obj = TypeVar("_T_obj", bound=ObjBase)


def property_obj(offset: int, cls: type[_T_obj], doc: str) -> _OffsetProp[_T_obj]:
    def _get(self: ObjBase) -> _T_obj:
        return cls(self.controller.read_i32(self.base_ptr + offset), self.controller)

    def _set(self: ObjBase, value: _T_obj) -> None:
        if self.controller != value.controller:
            raise ValueError("cannot assign an object from another controller")
        self.controller.write_i32(value.base_ptr, self.base_ptr + offset)

    return OffsetProperty(_get, _set, None, f"{cls.__name__}: {doc}", offset)


class ObjId(ObjBase):
    """
    ObjNode 对象末尾的(index, rank)对象

    游戏内用于 ObjNode 的识别和储存.
    """

    OBJ_SIZE = 4

    index = property_u16(0, "对象索引")

    rank = property_u16(2, "对象序列号")

    def __eq__(self, val: Self | tuple[int, int]) -> bool:
        """
        ObjId 比较相等 与其他 ObjId 比较或与(index, rank)比较

        Args:
            val: 另一个 ObjId 对象或(index, rank)一样的可解包对象
        Returns:
            "表示相同对象"返回 True
        """
        if isinstance(val, ObjId):
            return ((self.controller.read_u32(self.base_ptr) ==
                     val.controller.read_u32(val.base_ptr))
                    and self.controller == val.controller)
        index, rank = val
        return self.controller.read_u32(self.base_ptr) == ((rank << 16) | index)

    def __ne__(self, val: Self | tuple[int, int]) -> bool:
        return not self.__eq__(val)

    def __str__(self) -> str:
        return f"(index={self.index}, rank={self.rank})"

    def tpl(self) -> tuple[int, int]:
        """
        Returns:
            (index, rank)元组
        """
        return self.index, self.rank


class ObjNode(ObjBase, abc.ABC):
    """
    Plant Zombie 等等, 在 pvz 中由 ObjList 数组进行内存管理的对象

    Attributes:
        id: ObjNode 对象末尾的 ObjId 对象
    """
    __slots__ = ("id",)

    def __init__(self, base_ptr: int, ctler: Controller) -> None:
        super().__init__(base_ptr, ctler)
        self.id = ObjId(base_ptr + self.OBJ_SIZE - 4, ctler)

    ITERATOR_FUNC_ADDRESS: ClassVar[int] = NotImplemented
    """返回 pvz 中迭代对象的函数地址, 必须在所有非抽象子类中赋值"""

    ITERATOR_P_BOARD_REG: ClassVar[str] = "edx"
    """迭代对象函数用于存储 Board 指针的寄存器, reanimation 和粒子系统为 eax, 其他为 edx"""

    is_dead: OffsetProperty = NotImplemented
    """对象是否存活, 必须在所有非抽象子类中赋值"""


_T_node = TypeVar("_T_node", bound=ObjNode)


class ObjList(ObjBase, Sequence[_T_node], abc.ABC):
    """
    游戏中管理各类对象内存的数组, 即函数表中 DataArray 对象

    仅帮助 type hint 用, 请勿直接使用, 而是使用 obj_list 函数构造.
    """
    OBJ_SIZE = 28

    max_length = property_i32(4, "最大时对象数")

    next_index = property_i32(12, "下一个对象的索引")

    obj_num = property_i32(16, "当前对象数量")

    next_rank = property_i32(20, "下一个对象的序列号")

    def __len__(self) -> int:
        """
        返回最大时对象数, 与 max_length 相同

        Returns:
            最大时对象数
        """
        return self.controller.read_i32(self.base_ptr + 4)

    def at(self, index: int) -> _T_node:
        """
        返回 index 对应下标的元素

        Args:
            index: 非负索引, 不做任何检查
        Returns:
            对应下标的元素, 不确保存活
        """

    @overload
    def __getitem__(self, index: SupportsIndex) -> _T_node:
        """
        返回 index 对应下标的元素

        Args:
            index: 整数索引, 即支持负数索引
        Returns:
            对应下标的元素, 不确保存活
        Raises:
            IndexError: 若 index 超出范围抛出
        """

    @overload
    def __getitem__(self, index: slice) -> list[_T_node]:
        """
        返回 slice 切片对应的列表

        若在遍历返回值的时候有新对象生成或死亡, 该方法没法动态调整.

        Args:
            index: 切片索引
        Returns:
            对应切片的列表, 不保证其中任何成员存活
        """

    def __getitem__(self, index): ...

    def __invert__(self) -> Iterator[_T_node]:
        """
        迭代所有未回收对象的迭代器, 利用原版函数在迭代过程中动态寻找下一个对象
        
        Returns:
            迭代器, 仅迭代存活对象
        Examples:
            >>> l: ObjList = ...
            >>> for objects in ~l:
            ...     ...  # do something
            迭代l中所有未回收的对象
        """

    @property
    def alive_iterator(self) -> Iterator[_T_node]:
        """与 __invert__() 相同"""
        return ~self

    @overload
    def find(self, index: SupportsIndex | ObjId, /) -> _T_node | None:
        """
        通过 index 查找对象

        用 SupportsIndex 查找时, 未回收对象返回 T.
        用 ObjId 查找时, 在对应 index 位置对象 rank 相同时返回 T.

        Args:
            index: 整数索引, ObjId 对象或(index, rank)可解包对象
        Returns:
            存在未回收的对应对象返回, 否则返回 None.
        Raises:
            TypeError: index 不是 int 和 ObjId
        Example:
            >>> self.find(-1)
            当前最后一个对象(len(self) - 1)未回收时返回, 否则返回 None
        """

    @overload
    def find(self, idx: int, rank: int, /) -> _T_node | None:
        """
        通过(index, rank)组查找对象

        Args:
            idx: 整数索引
            rank: 序列号
        Returns:
            存在对应对象返回, 否则返回 None.
        Example:
            >>> self.find(1, 1)
            若 idx == 1 的对象的 rank==1, 返回该对象, 否则返回 None
        """

    def find(self, *args): ...

    def reset_stack(self) -> Self:
        """
        清栈.

        在所有对象都被回收时调用. 让本对象之后申请对象从0开始申请

        Returns:
            self
        Raises:
            PvzStatusError: 当有对象未回收时抛出
        """

    @abc.abstractmethod
    def free_all(self) -> Self:  # DataArrayFreeAll会泄露动画对象.
        """
        删除存活的所有对象.

        Returns:
            self
        """

    def set_next_idx(self, idx: int) -> Self:
        """
        设置下一个对象的编号, 若 idx 大于当前最大长度, 会调整最大长度至和 idx 相同.

        当前实现: 为将 idx 和 next_idx 在"栈位"对应位置中交换.
        "调整长度"当前实现: 从所需最高到当前长度倒序添加; 与 ize 一开始类似且调整长度后无需再次交换.

        Args:
            idx: 下一个对象的编号
        Returns:
            self
        Raises:
            ValueError: idx 不合法或 idx 所在对象未回收时抛出.
        """


def obj_list(node_cls: type[_T_node]) -> type[ObjList[_T_node]]:
    """
    根据 node_cls 构造对应的 NodeClsObject 的父类
    
    Args:
        node_cls: ObjNode 的子类
    Returns:
        管理 node_cls 对象的数组的父类
    """

    class _ObjIterator(Iterator[_T_node]):
        def __init__(self, ctler: Controller, _iterate_func_asm):
            self._current_ptr = 0
            self._controller = ctler
            self._iterate_func_asm = _iterate_func_asm

        def __next__(self) -> _T_node:
            self._controller.result_u64 = self._current_ptr
            self._controller.run_code(self._iterate_func_asm)
            if (self._controller.result_u64 >> 32) == 0:
                raise StopIteration
            self._current_ptr = self._controller.result_u32
            return node_cls(self._current_ptr, self._controller)

        def __iter__(self) -> Self:
            return self

    class _ObjListImplement(ObjList[_T_node], abc.ABC):
        def __init__(self, base_ptr: int, ctler: Controller):
            super().__init__(base_ptr, ctler)
            self._array_base_ptr = ctler.read_u32(base_ptr)
            self._code = f"""
                push esi
                mov esi, [0x6a9ec0]
                mov {node_cls.ITERATOR_P_BOARD_REG}, [esi + 0x768]
                mov esi, {self.controller.result_address}
                call {node_cls.ITERATOR_FUNC_ADDRESS}
                mov [esi + 4], al
                pop esi
                ret"""  # 可恶的reg优化
            self._iterate_func_asm = None

        def at(self, index: int) -> _T_node:
            return node_cls(self._array_base_ptr + node_cls.OBJ_SIZE * index, self.controller)

        def find(self, *args) -> _T_node | None:
            match args:
                case (index, rank):  # index: int; rank: int
                    target = self.at(index)
                    # try:  # unbearable slow...
                    #     target = self[index]
                    # except IndexError:
                    #     return None
                    return target if target.id.rank == rank else None
                case (idx, ):
                    if isinstance(idx, SupportsIndex):
                        target = self.at(idx.__index__())
                        return target if target.id.rank != 0 else None
                    if isinstance(idx, ObjId):
                        target = self.at(idx.index)
                        return target if target.id == idx else None
                    raise TypeError("index must be int or ObjId instance")
                case _:
                    raise ValueError("the function should have 1 or 2 parameters, "
                                     f"not {len(args)} parameters")

        def __getitem__(self, index: SupportsIndex | slice):
            if isinstance(index, SupportsIndex):
                index = index.__index__()
                i = index if index >= 0 else index + len(self)
                if i >= len(self) or i < 0:
                    raise IndexError("sequence index out of range")
                return self.at(i)
            if isinstance(index, slice):
                start, stop, step = index.indices(len(self))
                return [self[i] for i in range(start, stop, step)]
            raise TypeError(f"index must be int or slice, not {type(index).__name__} instance")

        def __invert__(self):
            if self._iterate_func_asm is None:
                self._iterate_func_asm = asm.decode(self._code, self.controller.asm_address)
            return _ObjIterator(self.controller, self._iterate_func_asm)

        def reset_stack(self) -> Self:
            if self.obj_num:
                raise PvzStatusError(
                    f"cannot reset stack when there are still {self.obj_num} objects alive")
            next_idx = self.next_index
            self.next_index = 0
            length = self.max_length
            self.max_length = 0
            while next_idx != length:
                t = self.at(next_idx).id
                next_idx = t.index
                t.index = 0
            return self

        def _assert_size(self, size: int) -> Self:
            if size <= (current_len := self.max_length):
                return self
            if self.next_index == current_len:
                self.next_index = size
            else:
                for it in self:
                    if it.id.index == current_len:
                        it.id.index = size
                        break
            self.max_length = size
            node = self.at(current_len)
            node.id.index = self.next_index
            node.is_dead = True
            for i in range(current_len + 1, size):
                node = self.at(i)
                node.id.index = i - 1
                node.is_dead = True
            self.next_index = size - 1
            return self

        def set_next_idx(self, idx: int) -> Self:
            if idx < 0:
                raise ValueError(f"next index should be non-negative, not {idx}")
            if self.at(idx).id.rank != 0:
                raise ValueError(f"object at index {idx} is still unavailable")
            self._assert_size(idx + 1)
            if idx == self.next_index:
                return self
            target_node = self.at(idx)
            first_node = self.at(self.next_index)
            before_node = first_node
            while before_node.id.index != idx:
                before_node = self.at(before_node.id.index)
            before_node.id.index, self.next_index = self.next_index, idx
            target_node.id.index, first_node.id.index = first_node.id.index, target_node.id.index
            return self

    return _ObjListImplement
